package daikon.inv.quaternary.fourScalar;

import daikon.PptSlice;
import daikon.inv.Invariant;
import daikon.inv.InvariantStatus;
import daikon.inv.OutputFormat;

public class DivisionPattern extends FourScalar {

	private static/* @Prototype */DivisionPattern proto;

	/** Returns the prototype invariant for Reverse **/
	public static/* @Prototype */DivisionPattern get_proto() {
		if (proto == null)
			proto = new /* @Prototype */DivisionPattern();
		return proto;
	}

	// We are Serializable, so we specify a version to allow changes to
	// method signatures without breaking serialization. If you add or
	// remove fields, you should change this number to the current date.
	private static final long serialVersionUID = 20121008L;
	private static boolean dkconfig_enabled = true;

	protected DivisionPattern() {
		super();
	}

	public DivisionPattern(PptSlice slice) {
		super(slice);
	}

	@Override
	public InvariantStatus check_modified(long v1, long v2, long v3, long v4,
			int count) {

		if (v1 != v2 * v3 + v4) {
			return InvariantStatus.FALSIFIED;
		}

		return InvariantStatus.NO_CHANGE;
	}

	@Override
	public InvariantStatus add_modified(long v1, long v2, long v3, long v4,
			int count) {
		return check_modified(v1, v2, v3, v4, count);
	}

	@Override
	protected double computeConfidence() {
		return Invariant.conf_is_ge(ppt.num_values(), 5);
	}

	@Override
	public String format_using(OutputFormat format) {
		if (format.isJavaFamily())
			return format_java_family(format);

		if (format == OutputFormat.DAIKON)
			return format_daikon();
		if (format == OutputFormat.ESCJAVA)
			return format_esc();
		if (format == OutputFormat.SIMPLIFY)
			return format_simplify();
		if (format == OutputFormat.VCC)
			return format_vcc(format);
		if (format == OutputFormat.VCC_PLUS_JAVA)
			return format_vcc_plus_java(format);
		if (format == OutputFormat.ESCJ_PLUS_JAVA) { 
			return format_escj_plus_java(format);
		} 
		if (format == OutputFormat.SPECSHARP_PLUS_CSHARP) { 
			return format_specsharp_plus_csharp(format);
		} 

		return format_unimplemented(format);
	}

	private String format_vcc_plus_java(OutputFormat format) {
		return format_vcc(OutputFormat.VCC) + "->"
				+ format_using(OutputFormat.JAVA);
	}

	private String format_vcc(OutputFormat format) {
		String a = this.var1().name_using(OutputFormat.VCC);
		String b = this.var2().name_using(OutputFormat.VCC);
		String c = this.var3().name_using(OutputFormat.VCC);
		String d = this.var4().name_using(OutputFormat.VCC);
		String daikon_str = String.format("%s==%s*%s+%s", a, b, c, d);
		return this.getClass().getName() + ": " + daikon_str;
	}

	private String format_daikon() {
		String a = this.var1().name();
		String b = this.var2().name();
		String c = this.var3().name();
		String d = this.var4().name();
		String daikon_str = String.format("%s==%s*%s+%s", a, b, c, d);
		return daikon_str;
	}

	private String format_esc() {
		throw new UnsupportedOperationException(
				"ESC format not implemented for this invariant");
	}

	private String format_simplify() {
		throw new UnsupportedOperationException(
				"Simplify format not implemented for this invariant");
	}

	private String format_java_family(OutputFormat format) {
		String a = this.var1().name_using(format);
		String b = this.var2().name_using(format);
		String c = this.var3().name_using(format);
		String d = this.var4().name_using(format);
		String java_str = String.format("%s==%s*%s+%s", a, b, c, d);
		return java_str;
	}

	/** instantiates the invariant on the specified slice **/
	public DivisionPattern instantiate_dyn(PptSlice slice) /* @Prototype */{
		return new DivisionPattern(slice);
	}

	/** returns whether or not this invariant is enabled **/
	public boolean enabled() {
		return dkconfig_enabled;
	}

	protected Invariant resurrect_done(int[] permutation) {
		return this;
	}
}
