#! /usr/bin/python

from ginpink import read_methods, location, extension
import argparse
import jml2exec
from reprocess import separator

methods_with_Integer = ['Arrays.fillInteger0', 'Arrays.fillInteger1', 'Arrays.hashCodeInteger', 'Arrays.insertionSort0', 'Arrays.merge0']

# from reprocess import process_line, separator

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Check that all required invariants are there.')
    parser.add_argument('--tofile', dest='ext', action='store',
                        help='write results to separate files with extension %(dest)s')
    parser.add_argument('-N', dest='niter', action='store', type=int,  default=16,
                        help='expect %(dest)s iterations (default: %(default)s)')
    parser.add_argument('classes', nargs='+', metavar='C', action='append', \
                        help='a class (without extension) to be processed; file %(dest)s' + \
                            extension + ' must exist')
    args = parser.parse_args()
    classes = args.classes[0]
    methods = []
    for c in classes:
        methods += read_methods(c)
    for pm, post, invs, pres in methods:
        mn = pm.name
        cn = pm.enclosing.name
        pkg = pm.package
        ii = set()
        print "************ %s ************" % (cn + '.' + mn)
        for i in invs:
            instr = i.strip().rstrip(';')
            ex = jml2exec.ExecutableJML(i)
            ex.make_executable()
            extext = ex.get_text()
            if extext.find('QTArrays') != -1 or extext.find('QTLists') != -1:
                extext = extext.replace('QTArrays', 'java3.theories.QTArrays')
                extext = extext.replace('QTLists', 'java3.theories.QTLists')
            else:
                extext = extext.replace('TArrays', 'java3.theories.TArrays')
                extext = extext.replace('TLists', 'java3.theories.TLists')
            # inline = ex.get_text() + separator + i
            if cn == 'Arrays':
                isIA = False
                if cn + '.' + mn in methods_with_Integer:
                    isIA = True
                qu = jml2exec.ExecutableJML(instr)
                qu.make_quantification(IntegerArrays=isIA, preconditions=True)
                qutext = qu.node.text()
            else:
                qutext = instr
            outline = extext + separator + qutext + '\n'
            ii.add(outline.strip())
        found = 0
        for n in range(1, args.niter + 1):
            filename = '.'.join([pkg, cn, mn, "iteration_" + str(n), "mutations"])
            with open(filename) as f:
                for line in f.readlines():
                    if line.strip() in ii:
                        print filename + ":", line.strip()
                        found += 1
        if found != len(ii):
            print "--> Method " + cn + '.' + mn, "found", found, "missing", len(ii) - found, "invariants"
        print '\n'
