'''
Created on Jun 18, 2013

@author: galeotti
'''
import glob
import os
import sys

if __name__ == '__main__':

    if len(sys.argv) != 6:
        print("Usage create_makefile.py <JAVA5_HOME> <JAVA6_HOME> <CASE_STUDIES_DIR> <FROM_RUNS> <TO_RUNS>")
        print("Example: create_makefile.py /System/Library/Java/JavaVirtualMachines/1.6.0.jdk/Contents/Home /Users/galeotti/workspace/loopmate-escj/case_studies 10")
        sys.exit()
    
    java5_home = sys.argv[1]
    java6_home = sys.argv[2]
    case_studies_dir = sys.argv[3]
    from_runs = int(sys.argv[4])
    to_runs = int(sys.argv[5])
    
    out_str = "" + "\n"
    
    out_str += "JAVA5_HOME=" + java5_home + "\n"
    out_str += "JAVA6_HOME=" + java6_home + "\n"
    out_str += "JAVABIN=$(JAVA6_HOME)/bin" + "\n"
    out_str += "#/opt/jdk1.6.0_33/bin" + "\n"

    out_str += "JAVAC=$(JAVABIN)/javac" + "\n"
    out_str += "JAVA=$(JAVABIN)/java" + "\n"
    
    out_str += "export PATH := $(JAVABIN):$(PATH)" + "\n"
    out_str += "export JAVA_HOME := $(JAVA6_HOME)" + "\n"
    out_str += "export JAVA5_HOME := $(JAVA5_HOME)" + "\n"
    out_str += "export JAVA6_HOME := $(JAVA6_HOME)" + "\n"

    out_str += "## PATHS VARIABLES ##" + "\n"
    out_str += "PWD := $(shell pwd)" + "\n"

    out_str += "TOOLS=$(PWD)/tools" + "\n"
    out_str += "DAIKON_DIR=$(TOOLS)/loopmate-daikon" + "\n"
    out_str += "JUNIT_DIR=$(TOOLS)/junit" + "\n"

    out_str += "JUNIT_JAR=$(JUNIT_DIR)/junit-4.10.jar" + "\n"
    out_str += "DAIKON_JAR=$(DAIKON_DIR)/dist/daikon.jar" + "\n"

    out_str += "OUTPUT_DIR=$(PWD)/results" + "\n"

    out_str += "CASE_STUDIES_DIR=$(PWD)/case_studies" + "\n"
    
    
    
    out_str += "##================================ SETUP =================================##" + "\n"

    out_str += "# setup target to download and build and tools" + "\n"
    out_str += "setup: toolsSetUp" + "\n"

    out_str += "####################### CASE STUDIES SETUP  ########################" + "\n"

    out_str += "## compile case study" + "\n"
    out_str += "$(CASE_STUDIES_DIR)/%/build: $(CASE_STUDIES_DIR)" + "\n"
    out_str += "\t" + "cd $(CASE_STUDIES_DIR)/$* && ant" + "\n"

    out_str += "CS_SETUP=$(CASE_STUDIES:%=$(CASE_STUDIES_DIR)/%/build)" + "\n"

    out_str += "## download and build all the case studies" + "\n"
    out_str += "csSetUp: $(CASE_STUDIES_DIR) $(CS_SETUP)" + "\n"


    out_str += "######################## TOOLS SETUP ############" + "\n"

    out_str += "toolsSetUp: $(DAIKON_JAR) $(JUNIT_JAR)" + "\n"

    out_str += "## target to download JUnit if it is not available yet" + "\n"
    out_str += "$(JUNIT_JAR):" + "\n"
    out_str += "\t" + "@echo \"** Downloading JUnit jar from SourceFourge.\"" + "\n"
    out_str += "\t" + "test -d $(TOOLS) || mkdir -p $(TOOLS)" + "\n"
    out_str += "\t" + "mkdir -p $(JUNIT_DIR)" + "\n"
    out_str += "\t" + "cd $(JUNIT_DIR);\\" + "\n"
    out_str += "\t" + "wget http://downloads.sourceforge.net/project/junit/junit/4.10/junit-4.10.jar" + "\n"

    out_str += "## build daikon jar files" + "\n"
    out_str += "$(DAIKON_JAR):$(DAIKON_DIR)" + "\n"
    out_str += "\t" + "@echo \"** Building Daikon jar files.\"" + "\n"
    out_str += "\t" + "cd $(DAIKON_DIR); \\" + "\n"
    out_str += "\t" + "ant" + "\n"

    out_str += "## target to clone daikon if it is not available yet" + "\n"
    out_str += "$(DAIKON_DIR):" + "\n"
    out_str += "\t" + "@echo \"** Downloading Daikon from repository.\"" + "\n"
    out_str += "\t" + "test -d $(TOOLS) || mkdir -p $(TOOLS)" + "\n"
    out_str += "\t" + "hg clone --insecure \\" + "\n"
    out_str += "\t\t" + "https://hg.st.cs.uni-saarland.de/hg/loopmate-daikon \\" + "\n"
    out_str += "\t\t" + "$@" + "\n"

    for case_study_dir in glob.glob(case_studies_dir + "/*"):
        case_study = os.path.basename(case_study_dir)
        methods_filename = case_study_dir + "/" + "methods.txt"
        case_study_targets = list()
        if  os.path.exists(methods_filename):
            class_targets = dict()
            for classname_and_methodname in open(methods_filename, 'r').readlines():
                classname_and_methodname_pair = classname_and_methodname.strip().split("..")
                classname = classname_and_methodname_pair[0]
                methodname = classname_and_methodname_pair[1]
                
                method_runs = list()
                for run_id in range(from_runs, to_runs + 1):
                    print("Case Study: " + case_study)
                    print("Class: " + classname)
                    print("Method: " + methodname)
                    print("Run :" + str(run_id))
                    run_id = "run_" + str(run_id)
                
                    out_dir = "$(OUTPUT_DIR)/" + case_study + "/" + classname + "/" + methodname + "/" + run_id
                    out_filename = out_dir + "/" + "loopmate-out.txt"
                    err_filename = out_dir + "/" + "loopmate-err.txt"

                    out_str += "\n"
                    out_str += "\n"
                    out_str += out_dir + ": $(CASE_STUDIES_DIR)/" + case_study + "/build $(EVOSUITE_JAR) $(DAIKON_JAR) $(JUNIT_JAR)" + "\n"
                    out_str += "\t" + "test -d " + out_dir + " || mkdir -p " + out_dir + "\n"
                    out_str += "\t" + "@echo \"** Running loop mate on $$@\"" + "\n"
                    out_str += "\t" + "python3.2 scripts/loopmate_escj.py \\" + "\n"
                    out_str += "\t\t" + "--sourcepath $(CASE_STUDIES_DIR)/" + case_study + "/src \\" + "\n"
                    out_str += "\t\t" + "--classpath $(CASE_STUDIES_DIR)/" + case_study + "/build/classes \\" + "\n"
                    out_str += "\t\t" + "--classname " + classname + " \\" + "\n"
                    out_str += "\t\t" + "--methodname " + methodname + " \\" + "\n"
                        
                    out_str += "\t\t" + "--target " + out_dir + " 2> " + err_filename + " > " + out_filename + "\n"

                    if not classname in class_targets.keys():
                        class_targets[classname] = list()
                    
                    method_runs.append(out_dir)
                    class_targets[classname].append(out_dir)
                    case_study_targets.append(out_dir)
            
                out_str += methodname + ": " + " ".join(method_runs) + "\n"
            
            for classname in class_targets.keys():
                out_str += "\n"
                out_str += classname + ": " + " ".join(class_targets[classname]) + "\n"
                
            out_str += "\n"
            out_str += case_study + ": " + " ".join(case_study_targets) + "\n"
                    
    makefile = open("makefile", "w")
    makefile.write(out_str)
    makefile.close()
    print ("New Makefile was generated.")
