from dynamate import print_iteration, get_iteration_name, \
    execute_chicory, ensure_dir, read_ginpink_mutations, execute_ginpink, \
    remove_escj_tautologies, merge_ginpink_invariants, prepare_evosuite, \
    execute_evosuite, instrument_invariants, prepare_daikon, execute_daikon, \
    LoopmateConfig, print_java_invariants_to_file, \
    print_daikon_invariants_to_file, prepare_escj, execute_escj, \
     ESCJResult, print_jml_invs_file, \
    execute_houdini, \
    write_verified_daikon_invariants_to_file, print_jml_invs_dict, \
    has_more_ginpink_mutations, isFixPoint, remove_evosuite_tautologies, Stats, \
    read_ginpink_invariants, number_of_invariants, is_empty, VerifiedInvariants, \
    read_invs_escj_plus_java, write_verified_ginpink_invariants_to_file, \
    EvoSuiteRestartLimit, write_escj_file, replace_in_file, \
    print_jml_invariants_to_file, GinpinkInvariant, write_ginpink_invariants, \
    GINPINK_DIR, NO_TIMEOUT, run_subprocess_timeout
import os
import datetime
import shutil
        
#==========================================================================================
        
def ginpink_stop_criteria(escj_output_str, daikon_inv_filenames, ginpink_inv_filenames, curr_ginpink_mutation):
    if escj_output_str != None and ESCJResult(escj_output_str).is_success():
        print("Stop criteria: ESCJ verified source code with loop invariants")
        return True
    
    if isFixPoint(daikon_inv_filenames, ginpink_inv_filenames):
        if escj_output_str != None and ((ESCJResult(escj_output_str).is_timeout() or ESCJResult(escj_output_str).is_vc_too_big()) and not ESCJResult(escj_output_str).is_loop_inv_fail()):
            print("Fix-point reached and ESCJ2 TIMEOUT or VC too big")  
            print("Stop criteria: unable to falsify candidates and ESCJ TIMEOUT (or VC too big). ESCJ capabilities are exceded.")
            print_jml_invs_dict()
            return True
        elif not has_more_ginpink_mutations(curr_ginpink_mutation): 
            print("Fix-point reached and no more ginpink candidates to add")  
            print("Stop criteria: daikon inferred invariants did not change and no new ginpink candidates to evaluate.")
            print_jml_invs_dict()
            return True
    
    print("Stop criteria was not met")
    return False

#==========================================================================================

def check_escj_candidate(work_dir, \
                         daikon_invariants, \
                         ginpink_invariants, \
                         candidate_loop_id=None, \
                         candidate_jml_string=None):

    loop_invariants = dict() 
    for loop_header in daikon_invariants.keys():
        if not loop_header.loop_id in loop_invariants.keys():
            loop_invariants[loop_header.loop_id] = list()
        
        daikon_invariants_set = daikon_invariants[loop_header]
        for daikon_invariant in daikon_invariants_set:
            jml_string = daikon_invariant[0].split("::")[1].strip()
            jml_annotation = "  //@loop_invariant " + jml_string + ";" + " //@nowarn"
            loop_invariants[loop_header.loop_id].append(jml_annotation)

    for loop_header in ginpink_invariants.keys():
        if not loop_header.loop_id in loop_invariants.keys():
            loop_invariants[loop_header.loop_id] = list()
        
        ginpink_invariants_list = ginpink_invariants[loop_header]
        for ginpink_invariant in ginpink_invariants_list:
            jml_string = ginpink_invariant.jml_string
            jml_annotation = "  //@loop_invariant " + jml_string + ";" + " //@nowarn"
            loop_invariants[loop_header.loop_id].append(jml_annotation)

    if candidate_loop_id != None and candidate_jml_string != None:
        if not candidate_loop_id in loop_invariants.keys():
            loop_invariants[candidate_loop_id] = list()
        candidate_jml_annotation = "  //@loop_invariant " + candidate_jml_string + ";" 
        loop_invariants[candidate_loop_id].append(candidate_jml_annotation)
        print("Creating JML input file for single check of candidate " + candidate_jml_string)
    else:
        print("Checking correctness assuming all candidates are proven")
        
    
    target_class_name = LoopmateConfig.target_class_name
    jml_template_file = LoopmateConfig.target_escj_source_path + "/" + target_class_name.replace(".", "/") + ".java"
    
    escj_filename = work_dir + "/" + "src" + "/" + target_class_name.replace(".", "/") + ".java"
    dest_dir = os.path.dirname(escj_filename)
    ensure_dir(dest_dir)
    
    print("Copying JML template file to  " + dest_dir)
    shutil.copy(jml_template_file, dest_dir)

    for loop_id in loop_invariants.keys():
        loop_invariant_list = loop_invariants[loop_id]
        pattern = "//$$" + loop_id + "$$"
        subst = "\n".join(loop_invariant_list)
        replace_in_file(escj_filename, pattern, subst)

    if LoopmateConfig.use_frame_loop_invariants == True:
        print("Enabling frame loop invariants (if any)")
        pattern = "//frame_loop_invariant "
        subst = "//@loop_invariant "
        replace_in_file(escj_filename, pattern, subst)
    
    if candidate_loop_id == None and candidate_jml_string == None:
        no_post_warnings = False
    else:
        no_post_warnings = True
    
    escj_output_filename, escj_output_str = execute_escj(work_dir, \
                                                         class_to_check=LoopmateConfig.target_class_name, \
                                                         routine_to_check=LoopmateConfig.target_method_name, \
                                                         only_loop_inv_check=no_post_warnings)

    return escj_output_str

#==========================================================================================
def filter_daikon_invariants(daikon_candidates):
    daikon_invariants = dict()
    for loop_header in daikon_candidates.keys():
        candidates_set = set()
        daikon_invariants[loop_header] = set()
        for daikon_candidate in daikon_candidates[loop_header]:
            jml_part = daikon_candidate[0].split("::")[1]
            if VerifiedInvariants.is_verified_invariant(loop_header.loop_id, jml_part):
                print("Skipping verifying loop invariant " + jml_part)
                daikon_invariants[loop_header].add(daikon_candidate)
            else:
                candidates_set.add(daikon_candidate)
                
        daikon_candidates[loop_header] = candidates_set
    
    return daikon_invariants, daikon_candidates
#==========================================================================================
def filter_ginpink_invariants(ginpink_candidates):
    ginpink_invariants = dict()
    
    for loop_header in ginpink_candidates.keys():
        candidates_list = list() 
        ginpink_invariants[loop_header] = list()
        for ginpink_invariant in ginpink_candidates[loop_header]:
            if VerifiedInvariants.is_verified_invariant(loop_header.loop_id, ginpink_invariant.jml_string):
                print("Skipping verifying loop invariant " + ginpink_invariant.jml_string)
                ginpink_invariants[loop_header].append(ginpink_invariant)
            else:
                candidates_list.append(ginpink_invariant)
        
        ginpink_candidates[loop_header] = candidates_list

    return ginpink_invariants, ginpink_candidates

#==========================================================================================
def get_free_daikon_invariants(daikon_invariants, work_list, except_index= -1):
    free_daikon_invariants = dict()
    for loop_header in daikon_invariants.keys():
        new_set = set(daikon_invariants[loop_header])
        free_daikon_invariants[loop_header] = new_set
        
    for index in range(len(work_list)):
        if index != except_index:
            loop_id, candidate = work_list[index]
            if isinstance(candidate, GinpinkInvariant):
                continue
            
            daikon_loop_header = None
            for loop_header in free_daikon_invariants.keys():
                if loop_header.loop_id == loop_id:
                    daikon_loop_header = loop_header
            
            free_daikon_invariants[daikon_loop_header].add(candidate)
        
    return free_daikon_invariants

#==========================================================================================
def get_free_ginpink_invariants(ginpink_invariants, work_list, except_index= -1):
    free_ginpink_invariants = dict()
    for loop_header in ginpink_invariants.keys():
        new_list = list(ginpink_invariants[loop_header])
        free_ginpink_invariants[loop_header] = new_list
        
    for index in range(len(work_list)):
        if index != except_index:
            loop_id, candidate = work_list[index]
            if not isinstance(candidate, GinpinkInvariant):
                continue
            
            ginpink_loop_header = None
            for loop_header in free_ginpink_invariants.keys():
                if loop_header.loop_id == loop_id:
                    ginpink_loop_header = loop_header
            
            free_ginpink_invariants[ginpink_loop_header].append(candidate)
        
    return free_ginpink_invariants
#==========================================================================================
def get_ginpink_invariants(loop_headers, work_list):
    ginpink_invariants = dict()
    for loop_header in loop_headers:
        ginpink_invariants[loop_header] = list()
        
    
    for loop_id, candidate in work_list:
        if not isinstance(candidate, GinpinkInvariant):
            continue
        
        ginpink_loop_header = None
        for loop_header in ginpink_invariants.keys():
            if loop_header.loop_id == loop_id:
                ginpink_loop_header = loop_header
                
        ginpink_invariants[ginpink_loop_header].append(candidate)
        
    return ginpink_invariants


#==========================================================================================
def get_daikon_invariants(loop_headers, work_list):
    daikon_invariants = dict()
    for loop_header in loop_headers:
        daikon_invariants[loop_header] = set()
        
    
    for loop_id, candidate in work_list:
        if isinstance(candidate, GinpinkInvariant):
            continue
        
        daikon_loop_header = None
        for loop_header in daikon_invariants.keys():
            if loop_header.loop_id == loop_id:
                daikon_loop_header = loop_header
                
        daikon_invariants[daikon_loop_header].add(candidate)
        
    return daikon_invariants

#==========================================================================================

def execute_top_down_escj(work_dir, daikon_inv_filename, ginpink_inv_filename):
    
    daikon_candidates = read_invs_escj_plus_java(daikon_inv_filename)
    ginpink_candidates = read_ginpink_invariants(ginpink_inv_filename)

    daikon_invariants, daikon_candidates = filter_daikon_invariants(daikon_candidates)
    ginpink_invariants, ginpink_candidates = filter_ginpink_invariants(ginpink_candidates)

    daikon_candidate_count = 0
    work_list = list()
    for loop_header in daikon_candidates.keys():
        for daikon_candidate in daikon_candidates[loop_header]:
            t = loop_header.loop_id, daikon_candidate
            work_list.append(t)
            daikon_candidate_count += 1
    
    ginpink_candidate_count = 0
    for loop_header in ginpink_candidates.keys():
        for ginpink_invariant in ginpink_candidates[loop_header]:
            t = loop_header.loop_id, ginpink_invariant
            work_list.append(t)
            ginpink_candidate_count += 1
            

    print("Initial number of daikon candidates=" + str(daikon_candidate_count))
    print("Initial number of ginpink candidates=" + str(ginpink_candidate_count))

    iter_counter = -1
    current_job = 0
    while current_job < len(work_list):
        print("Current worklist length= " + str(len(work_list)) + " Current job index=" + str(current_job))
        iter_counter += 1 
        iter_work_dir = work_dir + "." + str(iter_counter)
        ensure_dir(iter_work_dir)
        loop_id, candidate = work_list[current_job]
        if isinstance(candidate, GinpinkInvariant):
            jml_str = candidate.jml_string
        else:
            jml_str = candidate[0].split("::")[1]

        free_daikon_invariants = get_free_daikon_invariants(daikon_invariants, work_list, except_index=current_job)
        free_ginpink_invariants = get_free_ginpink_invariants(ginpink_invariants, work_list, except_index=current_job)
        
        escj_ret_val = check_escj_candidate(iter_work_dir, \
                                            free_daikon_invariants, \
                                            free_ginpink_invariants, \
                                            candidate_loop_id=loop_id, \
                                            candidate_jml_string=jml_str)
             
        passed = False
        if ESCJResult(escj_ret_val).is_loop_inv_fail():
            print("Ginpink candidate is invalid " + jml_str)
        elif ESCJResult(escj_ret_val).is_timeout():
            print("Timeout checking ginpink candidate " + jml_str)
        elif ESCJResult(escj_ret_val).is_vc_too_big():
            print("VC too big. Cannot check ginpink candidate " + jml_str)
        elif ESCJResult(escj_ret_val).is_error():
            print("ESCJ Error. Cannot check ginpink candidate " + jml_str)
        elif ESCJResult(escj_ret_val).is_prover_cc_limit():
            print("Prover CC limit. Cannot conclude correctness of ginpink candidate " + jml_str)
        else:
            print("Ginpink Candidate checking passed " + jml_str)
            passed = True
            
        if passed:
            print("Candidate " + jml_str + " passed using " + str(len(work_list)) + " nowarns")
            print("A total of " + str(current_job + 1) + " out of " + str(len(work_list)) + " candidates passed.")
            print("Candidate passed. Checking next candidate (if any)")
            current_job += 1
        else:
            print("Candidate " + jml_str + " dit not pass using " + str(len(work_list)) + " nowarns")
            print("Removing candidate from work_list and re-checking all")
            del work_list[current_job]
            current_job = 0
        
    if len(work_list) > 0:
        print("Some candidates were verified as new loop invariants")
        print("A total of " + str(len(work_list)) + " out of " + str(daikon_candidate_count + ginpink_candidate_count) + " original candidates were verified using " + str(iter_counter) + " ESC/Java2 calls.")
        
        iter_counter += 1
        last_iter_work_dir = work_dir + "." + str(iter_counter)
        
        free_daikon_invariants = get_free_daikon_invariants(daikon_invariants, work_list)
        free_ginpink_invariants = get_free_ginpink_invariants(ginpink_invariants, work_list)
        escj_ret_val = check_escj_candidate(last_iter_work_dir, \
                                            free_daikon_invariants, \
                                            free_ginpink_invariants)
        
        Stats.notify_new_verified_escj_output(escj_ret_val)
        
        return escj_ret_val, free_daikon_invariants, free_ginpink_invariants
    else:
        print("Could not prove any subset of candidates to be loop invariants")
        return None, dict(), dict()
        

def execute_bottom_up_escj(work_dir, daikon_inv_filename, ginpink_inv_filename):
    
    daikon_candidates = read_invs_escj_plus_java(daikon_inv_filename)
    ginpink_candidates = read_ginpink_invariants(ginpink_inv_filename)

    daikon_invariants, daikon_candidates = filter_daikon_invariants(daikon_candidates)
    ginpink_invariants, ginpink_candidates = filter_ginpink_invariants(ginpink_candidates)
    
    iter_counter = -1
    while True:
        changed = 0

        for loop_header in daikon_candidates.keys():
            candidates = set()
            for daikon_candidate in daikon_candidates[loop_header]:
                iter_counter += 1
                iter_work_dir = work_dir + "." + str(iter_counter)
                
                jml_string = daikon_candidate[0].split("::")[1]
                
                escj_ret_val = check_escj_candidate(iter_work_dir, \
                                                    daikon_invariants, \
                                                    ginpink_invariants, \
                                                    candidate_loop_id=loop_header.loop_id, \
                                                    candidate_jml_string=jml_string)
                
                if ESCJResult(escj_ret_val).is_success():
                    Stats.notify_new_verified_escj_output(escj_ret_val)
                    print("Program was verified while checking correctness of daikon candidate. Exiting")
                    return escj_ret_val, daikon_invariants, ginpink_invariants
                elif ESCJResult(escj_ret_val).is_loop_inv_fail():
                    print("Daikon candidate is invalid " + jml_string)
                    candidates.add(daikon_candidate)
                elif ESCJResult(escj_ret_val).is_timeout():
                    print("Timeout checking daikon candidate " + jml_string)
                    candidates.add(daikon_candidate)
                elif ESCJResult(escj_ret_val).is_vc_too_big():
                    print("VC too big. Cannot check daikon candidate " + jml_string)
                    candidates.add(daikon_candidate)
                elif ESCJResult(escj_ret_val).is_error():
                    print("ESCJ output error. Cannot check daikon candidate " + jml_string)
                    candidates.add(daikon_candidate)
                elif ESCJResult(escj_ret_val).is_prover_cc_limit():
                    print("Prover CC limit. Cannot conclude correctness of daikon candidate " + jml_string)
                    candidates.add(daikon_candidate)
                else:
                    Stats.notify_new_verified_escj_output(escj_ret_val)
                    print("Daikon Candidate proved as loop invariant " + jml_string)
                    daikon_invariants[loop_header].add(daikon_candidate)
                    changed += 1
                    
            daikon_candidates[loop_header] = candidates

        for loop_header in ginpink_candidates.keys():
            candidates = list()
            for ginpink_candidate in ginpink_candidates[loop_header]:
                iter_counter += 1
                iter_work_dir = work_dir + "." + str(iter_counter)
                jml_string = ginpink_candidate.jml_string
                escj_ret_val = check_escj_candidate(iter_work_dir, \
                                                    daikon_invariants, \
                                                    ginpink_invariants, \
                                                    candidate_loop_id=loop_header.loop_id, \
                                                    candidate_jml_string=jml_string)
                
                if ESCJResult(escj_ret_val).is_success():
                    print("Program was verified while checking correctness of ginpink candidate. Exiting")
                    return escj_ret_val, daikon_invariants, ginpink_invariants
                elif ESCJResult(escj_ret_val).is_loop_inv_fail():
                    print("Ginpink candidate is invalid " + jml_string)
                    candidates.append(ginpink_candidate)
                elif ESCJResult(escj_ret_val).is_timeout():
                    print("Timeout checking ginpink candidate " + jml_string)
                    candidates.append(ginpink_candidate)
                elif ESCJResult(escj_ret_val).is_vc_too_big():
                    print("VC too big. Cannot check ginpink candidate " + jml_string)
                    candidates.append(ginpink_candidate)
                elif ESCJResult(escj_ret_val).is_error():
                    print("ESCJ Error. Cannot check ginpink candidate " + jml_string)
                    candidates.append(ginpink_candidate)
                elif ESCJResult(escj_ret_val).is_prover_cc_limit():
                    print("Prover CC limit. Cannot conclude correctness of ginpink candidate " + jml_string)
                    candidates.append(ginpink_candidate)
                else:
                    print("Ginpink Candidate proved as loop invariant " + jml_string)
                    ginpink_invariants[loop_header].append(ginpink_candidate)
                    changed += 1
        
            ginpink_candidates[loop_header] = candidates
            
        if changed == 0:
            print("No new candidates were proven as invariants. Exiting.")
            break
        else:
            print(str(changed) + " candidates were proven as invariants. Continuing.")
            continue
        
    print("Program was not verified. Exiting")
    return None, daikon_invariants, ginpink_invariants

#==========================================================================================
def jml2exec(jml_string):
    old_python_path = os.environ["PYTHONPATH"]
    print("current PYTHONPATH=" + old_python_path)
    print("Cleaning PYTHONPATH for python2.7 execution")
    os.environ["PYTHONPATH"] = ""
    cmd = "python2.7 " + GINPINK_DIR + "/jml2exec.py dynamate " \
        + " --class " + LoopmateConfig.target_class_name \
        + " --method " + LoopmateConfig.target_method_name \
        + " \"" + jml_string + "\"" 
        
    print("Executing command: " + cmd)
    stdout, stderr, timeout_ret_code = run_subprocess_timeout(cmd, NO_TIMEOUT, use_shell=True)
    
    print("Restoring PYTHONPATH to " + old_python_path)
    os.environ["PYTHONPATH"] = old_python_path
    for line in stdout.split("\n"):
        if "SUCCESS:" in line:
            ret_val_jml_string = line.split("SUCCESS:")[1].strip()
            return ret_val_jml_string
        elif "FAILURE:" in line:
            print("Warning: failure unoldering expression " + jml_string)
            return jml_string

    return jml_string
    
    pass

#==========================================================================================

def read_new_ginpink_invariants(iterCount, curr_ginpink_mutation, ginpink_dtrace_filenames, daikon_inv_gz_filename):
    iterXXX = get_iteration_name(iterCount - 1)
    ginpink_work_dir = LoopmateConfig.work_dir + "/" + "ginpink-new" + "/" + iterXXX + "." + str(curr_ginpink_mutation)
            
    Stats.highestGinpinkWave = curr_ginpink_mutation
    ginpink_inv_filename = read_ginpink_mutations(ginpink_work_dir, curr_ginpink_mutation)

    print("Checking new candidates are found")
    start_time = datetime.datetime.now()            
    ginpink_inv_dict = read_ginpink_invariants(ginpink_inv_filename)
    end_time = datetime.datetime.now()
    print("Ginpink Wave:" + str(curr_ginpink_mutation) + " Ginpink new invariants after typechecking= " + str(number_of_invariants(ginpink_inv_dict)) + " (" + str(end_time - start_time) + ")")
    if is_empty(ginpink_inv_dict):
        return ginpink_inv_filename
    
    if LoopmateConfig.remove_method_calls_within_specs == True:
        for loop_header in ginpink_inv_dict.keys():
            for ginpink_invariant in ginpink_inv_dict[loop_header]:
                jml_string = ginpink_invariant.jml_string
                new_jml_string = jml2exec(jml_string)
                ginpink_invariant.jml_string = new_jml_string
                
        write_ginpink_invariants(ginpink_inv_dict, ginpink_inv_filename)
        

    print("Executing Ginpink Checker on new ginpink candidates")
    start_time = datetime.datetime.now()            
    ginpink_inv_filename = execute_ginpink(ginpink_work_dir, ginpink_inv_filename, ginpink_dtrace_filenames, iterCount)
    end_time = datetime.datetime.now()

    print("Checking new candidates were found using testsuite execution")            
    ginpink_inv_dict = read_ginpink_invariants(ginpink_inv_filename)
    print("Ginpink Wave:" + str(curr_ginpink_mutation) + " Ginpink new invariants after testsuite execution= " + str(number_of_invariants(ginpink_inv_dict)) + " (" + str(end_time - start_time) + ")")
    if is_empty(ginpink_inv_dict):
        return ginpink_inv_filename
            
    print("Filter out tautologies using ESC/Java2 (sound)")
    start_time = datetime.datetime.now()            
    escj_tauto_work_dir = LoopmateConfig.work_dir + "/" + "escj-tauto" + "/" + iterXXX + "." + str(curr_ginpink_mutation)
    ginpink_inv_escj_tauto = remove_escj_tautologies(escj_tauto_work_dir, ginpink_inv_filename, iterCount)
    end_time = datetime.datetime.now()

    print("Checking new candidates were found using ESC/Java2 removal")            
    ginpink_inv_dict = read_ginpink_invariants(ginpink_inv_escj_tauto)
    print("Ginpink Wave:" + str(curr_ginpink_mutation) + " Ginpink new invariants after ESC/Java2 removal= " + str(number_of_invariants(ginpink_inv_dict)) + " (" + str(end_time - start_time) + ")")
    if is_empty(ginpink_inv_dict):
        return ginpink_inv_escj_tauto

    if LoopmateConfig.evosuite_tauto_is_enabled == True:
        print("Filter out tautologies using EvoSuite (unsound)")
        start_time = datetime.datetime.now()            
        evosuite_tauto_work_dir = LoopmateConfig.work_dir + "/" + "evosuite-tauto" + "/" + iterXXX + "." + str(curr_ginpink_mutation)
        ginpink_inv_evosuite_tauto = remove_evosuite_tautologies(evosuite_tauto_work_dir, daikon_inv_gz_filename, ginpink_inv_escj_tauto, iterCount)
        ginpink_inv_dict = read_ginpink_invariants(ginpink_inv_evosuite_tauto)
        end_time = datetime.datetime.now()
        print("Ginpink Wave:" + str(curr_ginpink_mutation) + " Ginpink new invariants after EvoSuite removal= " + str(number_of_invariants(ginpink_inv_dict)) + " (" + str(end_time - start_time) + ")")
        return ginpink_inv_evosuite_tauto
    else:
        return ginpink_inv_escj_tauto

#==========================================================================================
def execute_houdini_escj(daikon_work_dir_name, escj_work_dir_name, iterCount, daikon_inv_gz_file, ginpink_inv_filenames):
    
    print("Preparing ESCJ")
    jml_invs_filename, daikon_invariants_dict, escj_filename = prepare_escj(daikon_work_dir_name, escj_work_dir_name, iterCount, daikon_inv_gz_file, ginpink_inv_filenames)

    print("Executing ESCJ")
    escj_output_filename, escj_output_str = execute_escj(escj_work_dir_name, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)
    
    print("Checking ESCJ result")
    if ESCJResult(escj_output_str).is_success():
        print("ESCJ could fully verified the program at first check")
        daikon_invariants = read_invs_escj_plus_java(jml_invs_filename)
        ginpink_invariants = read_ginpink_invariants(ginpink_inv_filenames[-1])
        Stats.notify_new_verified_escj_output(escj_output_str)

    elif ESCJResult(escj_output_str).is_loop_inv_fail():   
        print("ESCJ could not check some loop invariants at first check")
        original_escj_output_str = escj_output_str
        escj_output_str, all_daikon_invs_were_verified, daikon_invariants, ginpink_invariants = execute_houdini(escj_work_dir_name, \
                                                                                                                        daikon_invariants_dict, \
                                                                                                                        jml_invs_filename, \
                                                                                                                        ginpink_inv_filenames, \
                                                                                                                        original_escj_output_str)
    elif not ESCJResult(escj_output_str).is_timeout() and not ESCJResult(escj_output_str).is_vc_too_big():
        print("ESCJ could check all loop invariants at first check")
        daikon_invariants = read_invs_escj_plus_java(jml_invs_filename)
        ginpink_invariants = read_ginpink_invariants(ginpink_inv_filenames[-1])

    
    elif ESCJResult(escj_output_str).is_timeout():
        print("ESCJ first check led to TIMEOUT. Cannot apply houdini")
        daikon_invariants = dict()
        ginpink_invariants = dict()

    elif ESCJResult(escj_output_str).is_vc_too_big():
        print("ESCJ first check led to VC too big. Cannot apply houdini")
        daikon_invariants = dict()
        ginpink_invariants = dict()
    
        
    return escj_output_str, daikon_invariants, ginpink_invariants
#==========================================================================================

def execute_ginpink_loop(junit_file_names, \
                         daikon_bin_dirs, \
                         daikon_inv_gz_filenames, \
                         daikon_dtrace_filenames, \
                         daikon_inv_filenames, \
                         java_invs_filenames, \
                         ginpink_dtrace_filenames, \
                         jml_invs_filenames, \
                         escj_verified_outputs, \
                         iterCount):
    
    curr_ginpink_mutation = 0
    ginpink_inv_filenames = list()
    escj_output_str = None
    current_ginpink_inv_filename = None
    
    while not ginpink_stop_criteria(escj_output_str, daikon_inv_filenames, ginpink_inv_filenames, curr_ginpink_mutation):
        Stats.nrOfGinpinkIterations += 1
        iterCount += 1
    
        print_iteration(iterCount)
        iterXXX = get_iteration_name(iterCount)
        
        if isFixPoint(daikon_inv_filenames, ginpink_inv_filenames) or curr_ginpink_mutation == 0:
            
            new_ginpink_inv_dict = dict()
            while is_empty(new_ginpink_inv_dict) and has_more_ginpink_mutations(curr_ginpink_mutation):
                curr_ginpink_mutation += 1
                print("Reading new ginpink invariants (mutation=" + str(curr_ginpink_mutation) + ")")
                daikon_inv_gz_filename = daikon_inv_gz_filenames[-1]
                new_ginpink_inv_filename = read_new_ginpink_invariants(iterCount, curr_ginpink_mutation, ginpink_dtrace_filenames, daikon_inv_gz_filename)
                new_ginpink_inv_dict = read_ginpink_invariants(new_ginpink_inv_filename)
                if is_empty(new_ginpink_inv_dict):
                    print("No new ginpink invariants found from mutation iteration (muation=" + str(curr_ginpink_mutation) + ")")

            if  is_empty(new_ginpink_inv_dict):
                print("Could not read new ginpink candidates and no more ginpink waves to read")
                print("Exiting outer loop")
                break
            else:
                
                if current_ginpink_inv_filename == None:
                    current_ginpink_inv_filename = new_ginpink_inv_filename
                else:
                    print("Merging current ginpink invariants with previous ginpink invariants")
                    merge_dir_name = os.path.dirname(new_ginpink_inv_filename)
                    merged_ginpink_inv_filename = merge_ginpink_invariants(merge_dir_name, current_ginpink_inv_filename, new_ginpink_inv_filename)
                    current_ginpink_inv_filename = merged_ginpink_inv_filename
                
                print("Instrument target class with detected invariants")
                daikon_inv_filename = daikon_inv_filenames[-1]
                daikon_work_dir_name = os.path.dirname(daikon_inv_filename)
                ginpink_dir = os.path.dirname(current_ginpink_inv_filename)
                current_ginpink_inv_filename_list = list()
                current_ginpink_inv_filename_list.append(current_ginpink_inv_filename)
                src_target_java_file_name = instrument_invariants(daikon_work_dir_name, daikon_inv_gz_filename, ginpink_dir, current_ginpink_inv_filename_list)
    
        while True:
            
            print("Preparing EvoSuite")
            evosuite_work_dir_name = LoopmateConfig.work_dir + "/" + "evosuite" + "/" + iterXXX;
            prepare_evosuite(src_target_java_file_name, evosuite_work_dir_name)
            
            restart_count = 0
            while restart_count < LoopmateConfig.max_number_of_evosuite_restarts_in_single_iteration:
                print("Executing EvoSuite")
                junit_file_name = execute_evosuite(src_target_java_file_name, evosuite_work_dir_name, iterCount)
                junit_file_names.append(junit_file_name)
            
                print("Preparing Daikon")
                daikon_work_dir_name = LoopmateConfig.work_dir + "/" + "daikon" + "/" + iterXXX
                prepare_daikon(daikon_work_dir_name, junit_file_names, iterCount)
            
                print("Executing Chicory for Daikon (std-visibility) and purity filename (if any)")
                daikon_dir_bin = daikon_work_dir_name + "/" + "bin"
            
                target_class_name = LoopmateConfig.target_class_name
                purity_filename = LoopmateConfig.target_java_source_path + "/" + target_class_name.replace(".", "/") + "_" + LoopmateConfig.target_method_name + ".pure"
                if os.path.exists(purity_filename):
                    print("Purity file for Daikon was found: " + purity_filename)
                    daikon_dtrace_file = execute_chicory(daikon_work_dir_name, daikon_dir_bin, junit_file_name, True, purity_filename)
                else:
                    print("No purity file was found for Daikon. We looked for " + purity_filename)
                    daikon_dtrace_file = execute_chicory(daikon_work_dir_name, daikon_dir_bin, junit_file_name, True, None)
                        
                print("Executing Daikon")
                daikon_dtrace_file, daikon_inv_gz_file = execute_daikon(daikon_work_dir_name, daikon_dtrace_file, daikon_dtrace_filenames, junit_file_name)
    
                if daikon_inv_gz_file != None:
                    break
                else:
                    restart_count += 1
    
            if daikon_inv_gz_file == None:
                raise EvoSuiteRestartLimit()
            
            daikon_dtrace_filenames.append(daikon_dtrace_file)
            daikon_inv_gz_filenames.append(daikon_inv_gz_file)
            
            daikon_bin_dir = daikon_work_dir_name + "/" + "bin"
            daikon_bin_dirs.append(daikon_bin_dir)
    
            print("Saving Java invariants to a file")
            java_invs_filename = daikon_work_dir_name + "/" + "java_invs" + str(iterCount) + ".txt"
            print_java_invariants_to_file(daikon_work_dir_name, java_invs_filename, daikon_inv_gz_file)
            java_invs_filenames.append(java_invs_filename)
                    
            print("Saving Daikon invariants to a file")
            daikon_invs_filename = daikon_work_dir_name + "/" + "daikon_invs" + str(iterCount) + ".txt"
            print_daikon_invariants_to_file(daikon_invs_filename, daikon_inv_gz_file)
            daikon_inv_filenames.append(daikon_invs_filename)
            
            print("Executing Chicory for GinPink (no std-visibility) and no purity filename")
            ginpink_work_dir_name = LoopmateConfig.work_dir + "/" + "ginpink" + "/" + iterXXX
            ensure_dir(ginpink_work_dir_name)
            ginpink_dtrace_file = execute_chicory(ginpink_work_dir_name, daikon_dir_bin, junit_file_name, False, None)
            ginpink_dtrace_filenames.append(ginpink_dtrace_file)
                    
            print("Executing Ginpink Checker on new ginpink candidates")
            last_ginpink_dtrace_list = list()
            last_ginpink_dtrace_list.append(ginpink_dtrace_filenames[-1])
            current_ginpink_inv_filename = execute_ginpink(ginpink_work_dir_name, current_ginpink_inv_filename, last_ginpink_dtrace_list, iterCount)
            ginpink_inv_filenames.append(current_ginpink_inv_filename)

            print("Instrument target class with detected invariants")
            src_target_java_file_name = instrument_invariants(daikon_work_dir_name, daikon_inv_gz_file, ginpink_work_dir_name, ginpink_inv_filenames)

            if len(ginpink_inv_filenames) > 1 and isFixPoint(daikon_inv_filenames, ginpink_inv_filenames):
                print("Fixpoint found. Cannot remove more candidates. Executing ESCJ now")
                break
            else:
                Stats.nrOfGinpinkIterations += 1
                iterCount += 1
                print_iteration(iterCount)
                iterXXX = get_iteration_name(iterCount)                
            
        print("Preparing ESCJ")
        escj_work_dir_name = LoopmateConfig.work_dir + "/" + "escj" + "/" + iterXXX
        ensure_dir(escj_work_dir_name)
        escj_plus_java_filename = escj_work_dir_name + "/" + "daikon_invs" + str(iterCount) + ".txt"
        print_jml_invariants_to_file(escj_plus_java_filename, daikon_inv_gz_filename)
        
        escj_output_str, daikon_invariants, ginpink_invariants = execute_houdini_escj(daikon_work_dir_name, escj_work_dir_name, iterCount, daikon_inv_gz_file, ginpink_inv_filenames)
#        escj_output_str, daikon_invariants, ginpink_invariants = execute_top_down_escj(escj_work_dir_name, escj_plus_java_filename, current_ginpink_inv_filename)

        if not VerifiedInvariants.all_daikon_invariants_are_verified(daikon_invariants):
            VerifiedInvariants.add_new_verified_daikon_invariants(daikon_invariants)
            verified_loop_invariants_filename = write_verified_daikon_invariants_to_file(daikon_invariants, escj_work_dir_name, iterCount)
            VerifiedInvariants.add_verified_daikon_invariants_filename(verified_loop_invariants_filename)
            Stats.notify_new_verified_escj_output(escj_output_str)

        if not VerifiedInvariants.all_ginpink_invariants_are_verified(ginpink_invariants):
            VerifiedInvariants.add_new_verified_ginpink_invariants(ginpink_invariants)
            verified_loop_invariants_filename = write_verified_ginpink_invariants_to_file(ginpink_invariants, escj_work_dir_name, iterCount)
            VerifiedInvariants.add_verified_ginpink_invariants_filename(verified_loop_invariants_filename)
            Stats.notify_new_verified_escj_output(escj_output_str)

        if escj_output_str != None and ESCJResult(escj_output_str).is_success():
            print("ESC/Java2 proved program specification fully!")
            print_jml_invs_dict()                            

        elif escj_output_str != None and ESCJResult(escj_output_str).is_timeout() and not ESCJResult(escj_output_str).is_loop_inv_fail():
            print("ESC/Java2 could not finish applying Houdini and fixpoint already reached")
            print("Exiting ginpink loop because new candidates hardly will alleviate the theorem prover.")
            break
