'''
Created on Sep 20, 2012

@author: galeotti
'''
import os
from dynamate import changeJVM, \
    print_java_version, clean_previous_dirs, execute_escj_first_time, \
    print_iteration, get_iteration_name, prepare_evosuite, execute_evosuite, \
    prepare_daikon, execute_chicory, EvoSuiteRestartLimit, Stats, \
    print_java_invariants_to_file, print_daikon_invariants_to_file, \
    print_java_invariants, ensure_dir, instrument_invariants, prepare_escj, \
    execute_escj, print_jml_invs_file, \
    print_jml_invs_dict, \
    write_verified_daikon_invariants_to_file, \
    execute_houdini, print_stats, ESCJError, \
    EvoSuiteError, GinpinkError, DaikonError, FileNotFoundError, \
    TrivialVerification, GinpinkTimeout, DaikonTimeout, execute_daikon, \
    read_invs_escj_plus_java, isFixPoint, VerifiedInvariants, \
    read_ginpink_invariants

from dynamate import ConfigurationError, LoopmateConfig, ESCJResult

from ginpink_loop import execute_ginpink_loop
import sys
import datetime
from optparse import OptionParser
import traceback

def daikon_stop_criteria(iterCount, escj_output_str, daikon_inv_filenames, ginpink_inv_filenames, all_daikon_invs_were_verified, curr_ginpink_mutation):
    if ESCJResult(escj_output_str).is_success():
        print("Stop criteria: ESCJ verified source code with loop invariants")
        return True
    
    if all_daikon_invs_were_verified:
        print("Fix-point reached")
        print("Stop criteria: all inferred daikon invariants were verified")
        return True
    
    if isFixPoint(daikon_inv_filenames, ginpink_inv_filenames) and not VerifiedInvariants.is_empty(): 
        print("Fix-point reached and some daikon loop invariants were verified")  
        print("Stop criteria: daikon inferred invariants did not change and verified loop invariants found")
        print_jml_invs_dict()
        return True
    
    print("Stop criteria was not met")
    return False

def execute_loopmate():
    
    print("LoopMate version 0.1")


    if not "JAVA6_HOME" in os.environ:
        print("Please set JAVA6_HOME environment variable")
        print("Loopmate is exiting because the JAVA6_HOME environment variable was not configured.")
        raise ConfigurationError()
    else:
        LoopmateConfig.java6_home = os.environ["JAVA6_HOME"]

    if not "JAVA5_HOME" in os.environ:
        print("Please set JAVA5_HOME environment variable")
        print("Loopmate is exiting because the JAVA5_HOME environment variable was not configured.")
        raise ConfigurationError()
    else:
        LoopmateConfig.java5_home = os.environ["JAVA5_HOME"]

    if not "PATH" in os.environ:
        LoopmateConfig.orig_path = None
    else:
        LoopmateConfig.orig_path = os.environ["PATH"]
        
    changeJVM(LoopmateConfig.java6_home)


    print_java_version()
    
    print("Cleaning previous Loopmate executions")
    clean_previous_dirs()

    tcn = LoopmateConfig.target_class_name
    src_target_java_file_name = LoopmateConfig.target_java_source_path + "/" + tcn.replace(".", "/") + ".java"

    iterCount = 0;
    escj_output_str = ""
    java_invs_filenames = list()
    jml_invs_filenames = list()
    daikon_dtrace_filenames = list()
    ginpink_dtrace_filenames = list()
    ginpink_inv_filenames = list()
    daikon_inv_filenames = list()
    daikon_inv_gz_filenames = list()
    junit_file_names = list()
    daikon_bin_dirs = list()
    escj_verified_outputs = list()

    all_daikon_invs_were_verified = False
    current_evosuite_restarts = 0
    
    execute_escj_first_time()
    
    curr_ginpink_mutation = 0
    
    while True:
        
        print_iteration(iterCount)
        iterXXX = get_iteration_name(iterCount)
        
        print("Preparing EvoSuite")
        evosuite_work_dir_name = LoopmateConfig.work_dir + "/" + "evosuite" + "/" + iterXXX;
        prepare_evosuite(src_target_java_file_name, evosuite_work_dir_name)
    
        print("Executing EvoSuite")
        junit_file_name = execute_evosuite(src_target_java_file_name, evosuite_work_dir_name, iterCount)
        if junit_file_name != None:
            junit_file_names.append(junit_file_name)
            
        print("Preparing Daikon")
        daikon_work_dir_name = LoopmateConfig.work_dir + "/" + "daikon" + "/" + iterXXX
        prepare_daikon(daikon_work_dir_name, junit_file_names, iterCount)

        print("Executing Chicory for Daikon (std-visibility) and purity filename (if any)")
        curr_dir_bin = daikon_work_dir_name + "/" + "bin"
        
        tcn = LoopmateConfig.target_class_name
        purity_filename = LoopmateConfig.target_java_source_path + "/" + tcn.replace(".", "/") + "_" + LoopmateConfig.target_method_name + ".pure"
        if os.path.exists(purity_filename):
            print("Purity file for Daikon was found: " + purity_filename)
            daikon_dtrace_file = execute_chicory(daikon_work_dir_name, curr_dir_bin, junit_file_name, True, purity_filename)
        else:
            print("No purity file was found for Daikon. We looked for " + purity_filename)
            daikon_dtrace_file = execute_chicory(daikon_work_dir_name, curr_dir_bin, junit_file_name, True, None)
            
        print("Executing Daikon")
        daikon_dtrace_file, daikon_inv_gz_file = execute_daikon(daikon_work_dir_name, daikon_dtrace_file, daikon_dtrace_filenames, junit_file_name)
        if daikon_inv_gz_file == None:
            print ("Daikon: no program points were covered")
            
            if current_evosuite_restarts == LoopmateConfig.max_number_of_evosuite_restarts_in_single_iteration:
                print("Loopmate reached the maximum number of EvoSuite restarts per iteration: " + str(LoopmateConfig.max_number_of_evosuite_restarts_in_single_iteration))
                print("Loopmate is exiting because we reached the maximum number of EvoSuite seeds")
                raise EvoSuiteRestartLimit()
            else:
                Stats.nrOfRestartedIterations += 1
                current_evosuite_restarts += 1        
                print ("Restarting Iteration: " + str(iterCount))
                print("Current restart number is " + str(current_evosuite_restarts))
                continue  #
        
        
        
        daikon_dtrace_filenames.append(daikon_dtrace_file)
        daikon_inv_gz_filenames.append(daikon_inv_gz_file)

        daikon_bin_dir = daikon_work_dir_name + "/" + "bin"
        daikon_bin_dirs.append(daikon_bin_dir)
            
        print("Saving Java invariants to a file")
        java_invs_filename = daikon_work_dir_name + "/" + "java_invs" + str(iterCount) + ".txt"
        print_java_invariants_to_file(daikon_work_dir_name, java_invs_filename, daikon_inv_gz_file)
        java_invs_filenames.append(java_invs_filename)
            
        print("Saving Daikon invariants to a file")
        daikon_inv_filename = daikon_work_dir_name + "/" + "daikon_invs" + str(iterCount) + ".txt"
        print_daikon_invariants_to_file(daikon_inv_filename, daikon_inv_gz_file)
        daikon_inv_filenames.append(daikon_inv_filename)

        if LoopmateConfig.verbosity >= 2:
            print("Printing detected Java invariants")
            print_java_invariants(daikon_work_dir_name, daikon_inv_gz_file)

        print("Executing Chicory for GinPink (no std-visibility) and no purity filename")
        ginpink_chicory_work_dir_name = LoopmateConfig.work_dir + "/" + "ginpink" + "/" + iterXXX
        ensure_dir(ginpink_chicory_work_dir_name)
        ginpink_dtrace_file = execute_chicory(ginpink_chicory_work_dir_name, daikon_bin_dir, junit_file_name, False, None)
        ginpink_dtrace_filenames.append(ginpink_dtrace_file)
        
        print("Instrument target class with detected invariants")
        src_target_java_file_name = instrument_invariants(daikon_work_dir_name, daikon_inv_gz_file, None, ginpink_inv_filenames)

        print("Preparing ESCJ")
        escj_work_dir_name = LoopmateConfig.work_dir + "/" + "escj" + "/" + iterXXX
        jml_invs_filename, daikon_invariants_dict, escj_filename = prepare_escj(daikon_work_dir_name, escj_work_dir_name, iterCount, daikon_inv_gz_file, ginpink_inv_filenames)
        jml_invs_filenames.append(jml_invs_filename)

        print("Executing ESCJ")
        escj_output_filename, escj_output_str = execute_escj(escj_work_dir_name, class_to_check=LoopmateConfig.target_class_name, routine_to_check=LoopmateConfig.target_method_name)
        
        if ESCJResult(escj_output_str).is_success():
            last_jml_invs_filename = jml_invs_filenames[len(jml_invs_filenames) - 1]
            print_jml_invs_file(last_jml_invs_filename)            
        
        elif not ESCJResult(escj_output_str).is_timeout() and not ESCJResult(escj_output_str).is_vc_too_big():
            print("")
            if ESCJResult(escj_output_str).is_loop_inv_fail():   
                print("ESCJ could not check some loop invariants at first check")
                original_escj_output_str = escj_output_str
                escj_output_str, all_daikon_invs_were_verified, verified_daikon_invariants, verified_ginpink_invariants = execute_houdini(escj_work_dir_name, \
                                                                                                                                         daikon_invariants_dict, \
                                                                                                                                         jml_invs_filename, \
                                                                                                                                         ginpink_inv_filenames, \
                                                                                                                                         original_escj_output_str)
            else:
                print("ESCJ could check all loop invariants at first check")
                all_daikon_invs_were_verified = True
                verified_daikon_invariants = read_invs_escj_plus_java(jml_invs_filename)
                 
            if len(verified_daikon_invariants.keys()) > 0:
                verified_loop_invariants_filename = write_verified_daikon_invariants_to_file(verified_daikon_invariants, escj_work_dir_name, iterCount)
                VerifiedInvariants.add_new_verified_daikon_invariants(verified_daikon_invariants)
                VerifiedInvariants.add_verified_daikon_invariants_filename(verified_loop_invariants_filename)
                escj_verified_outputs.append(escj_output_str)
                Stats.notify_new_verified_escj_output(escj_output_str)
                    
            if ESCJResult(escj_output_str).is_success():
                print_jml_invs_dict()

        elif ESCJResult(escj_output_str).is_timeout():
            print("First verification led to TIMEOUT. Cannot apply HOUDINI")
            print("Continue until fixpoint")
            pass

        elif ESCJResult(escj_output_str).is_vc_too_big():
            print("First verification led VC TOO BIG. Cannot apply HOUDINI")
            print("Continue until fixpoint")
            pass

        Stats.nrOfDaikonIterations += 1        
        iterCount += 1
        current_evosuite_restarts = 0
        
        if daikon_stop_criteria(iterCount, escj_output_str, daikon_inv_filenames, ginpink_inv_filenames, all_daikon_invs_were_verified, curr_ginpink_mutation) and not VerifiedInvariants.is_empty():
            print("Stop Condition was met.")
            break 

    if not ESCJResult(escj_output_str).is_success() and LoopmateConfig.ginpink_is_enabled == True:
        print("Entering Ginpink+Daikon iterations")
        execute_ginpink_loop(junit_file_names, \
                         daikon_bin_dirs, \
                         daikon_inv_gz_filenames, \
                         daikon_dtrace_filenames, \
                         daikon_inv_filenames, \
                         java_invs_filenames, \
                         ginpink_dtrace_filenames, \
                         jml_invs_filenames, \
                         escj_verified_outputs, \
                         iterCount)
        escj_output_str = escj_verified_outputs[-1]

    if ESCJResult(escj_output_str).is_timeout() or ESCJResult(escj_output_str).is_vc_too_big():
        print("* Final Number of Warnings: Unknown (Unable to verify within timebound)")
       
    else:
        if len(escj_verified_outputs) > 0:
            final_escj_output_str = escj_verified_outputs[-1]
        else:
            final_escj_output_str = escj_output_str
        Stats.final_number_of_warnings = ESCJResult(final_escj_output_str).get_warnings()
        warnings_diff = Stats.final_number_of_warnings - Stats.initial_number_of_warnings

        if ESCJResult(final_escj_output_str).is_success():
            print("* Final Number of Warnings: 0 (Verification was successful)")
        else:
            print("* Final Number of Warnings: " + str(Stats.final_number_of_warnings) + " (Verification was unsuccessful)")
            if len(VerifiedInvariants.verified_invariants) == 0:
                print("No loop invariant was successfully verified")
                warnings_diff = 0

        if warnings_diff > 0:
            print("* Diff Final-Initial : +" + str(warnings_diff))
        else:
            print("* Diff Final-Initial : " + str(warnings_diff))

#=====================================================================================

class LoopmateExecutor:
    
    def parse_command_line_args(self, command_line_args):
 
        parser = OptionParser()
        parser.add_option("--sourcepath", action="store", type="string", dest="target_sourcepath", help="directories with source files of target class", metavar="DIR", default=None)
        parser.add_option("--classpath", action="store", type="string", dest="target_classpath", help="class search path of directories and zip/jar files", metavar="DIR", default=None)
        parser.add_option("--classname", action="store", type="string", dest="target_classname", help="target class for verification", metavar="CLASS_NAME", default=None)
        parser.add_option("--methodname", action="store", type="string", dest="target_methodname", help="target method for verification", metavar="METHOD_NAME", default=None)
        parser.add_option("--target", action="store", type="string", dest="target_work_dir", help="directory for output files", metavar="DIR", default=None)

        (options, args) = parser.parse_args(args=command_line_args)
    
        if len(args) > 1:
            print("Unknown arguments " + " ".join(args))
            parser.print_help()
            return False

        if options.target_methodname == None:
            print("Missing mandatory argument: methodname")
            parser.print_help()
            return False
        else:
            LoopmateConfig.target_method_name = options.target_methodname
                
        if options.target_sourcepath == None:
            print("Missing mandatory argument: sourcepath")
            parser.print_help()
            return False
        else:
            LoopmateConfig.target_java_source_path = options.target_sourcepath
            tjsp = LoopmateConfig.target_java_source_path
            LoopmateConfig.target_escj_source_path = "/".join(tjsp.split("/")[:-1]) + "/" + "escj-src"
    
        if   options.target_classpath == None:
            print("Missing mandatory argument: classpath")
            parser.print_help()
            return False
        else:
            LoopmateConfig.target_class_path = options.target_classpath
                
        if options.target_classname == None:
            print("Missing mandatory argument: classname")
            parser.print_help()
            return False
        else:
            LoopmateConfig.target_class_name = options.target_classname
        

        if options.target_work_dir == None:
            print("Missing mandatory argument: target")
            parser.print_help()
            return False
        else:
            LoopmateConfig.work_dir = options.target_work_dir
            
        return True
        
    
    def execute_from_python_command_line(self, command_line_args):
        
        parsing_ok = self.parse_command_line_args(command_line_args)
        
        if not parsing_ok:
            return False

        total_start_time = datetime.datetime.now()
        try:
            execute_loopmate()
            print("")
            print("NormalExit")
            print("")

        except DaikonTimeout:
            print("")
            print("LoopmateException")
            print("DaikonTimeout")
            print("")
        except GinpinkTimeout:
            print("")
            print("LoopmateException")
            print("GinpinkTimeout")
            print("")
        except EvoSuiteRestartLimit:
            print("")
            print("LoopmateException")
            print("EvoSuiteRestartLimit")
            print("")
        except TrivialVerification:
            print("")
            print("LoopmateException")
            print("TrivialVerification")
            print("")

        except FileNotFoundError:
            print("")
            print("LoopmateError")
            print("FileNotFoundError")
            print("")
        except DaikonError:
            print("")
            print("LoopmateError")
            print("DaikonError")
            print("")
        except GinpinkError:
            print("")
            print("LoopmateError")
            print("GinpinkError")
            print("")
        except ConfigurationError:
            print("")
            print("LoopmateError")
            print("ConfigurationError")
            print("")
        except EvoSuiteError:
            print("")
            print("LoopmateError")
            print("EvoSuiteError")
            print("")
        except ESCJError:
            print("")
            print("LoopmateError")
            print("ESCJError")
            print("")
        except Exception as ex:
            print("")
            print("LoopmateCrash")
            print(str(ex))
            traceback.print_exc()

        finally:
            total_end_time = datetime.datetime.now()
            Stats.total_time = total_end_time - total_start_time
            print_stats()

if __name__ == '__main__':
    xpire_executor = LoopmateExecutor();
    command_line_args = sys.argv[1:]
    xpire_executor.execute_from_python_command_line(command_line_args)

