#!/usr/bin/env python
'''
Created on Oct 13, 2012

@author: galeotti
'''

import os
import glob
import datetime
import sys

NORMAL_TERMINATION_VERIFICATION = "Stop criteria: ESCJ verified source code with loop invariants"
NORMAL_TERMINATION_FIXPOINT = "Stop criteria: daikon inferred invariants did not change"

#ROOT_WORK_DIR_NAME = "/Users/galeotti/workspace/loopmate-python/results"

CWD = os.getcwd()
PYTHON_BIN = "python3.2"
LOOPMATE_PY = CWD + "/scripts/loopmate.py"
#=======================================================================
class LoopmateExecution:
    number_of_iterations = 0
    number_of_unguided_iterations = 0
    total_time = 0
    evosuite_time = 0
    daikon_time = 0
    escj_time = 0
    execution_time_in_seconds =0
    number_of_escj_timeouts = 0
    number_of_evosuite_goals = None
    

#=======================================================================
class LoopmateStats:
    abnormal_termination = 0
    normal_termination = 0
    stop_criteria_fixpoint = 0
    stop_criteria_verification = 0
    number_of_repetitions = 0

    loopmate_executions = None
#=====================================================================================

if __name__ == '__main__':
    
    if len(sys.argv) != 2:
        print("Usage python stats.py <results_dir_name>")
        print("Example python stats.py /Users/galeotti/workspace/loopmate-python/results")
        sys.exit()
    
    root_work_dir_name = sys.argv[1]
    
    case_studies = list()


    wild_card = root_work_dir_name + "/" + "*" 
    print("Listing " + wild_card) 
    benchmarks = glob.glob(wild_card)
    for benchmark in benchmarks:
        benchmark_id = benchmark.split("/")[-1]
        print ("benchmark: " + benchmark_id)

        wild_card = root_work_dir_name + "/" + benchmark_id + "/*" 
        classnames = glob.glob(wild_card)
        for classname in classnames:
            classname_id = classname.split("/")[-1]
            print ("classname: " + classname_id)

            wild_card = root_work_dir_name + "/" + benchmark_id + "/" + classname_id + "/*" 
            methods = glob.glob(wild_card)
            for method in methods:
                method_id = method.split("/")[-1]
                print("method: " + method_id)
            
                case_studies.append(benchmark_id + "/" + classname_id + "/" + method_id)

    loopmate_runs = dict()
    for case_study in case_studies:
        wild_card = root_work_dir_name + "/" + case_study + "/" +"run_*" 
        print("Listing " + wild_card) 
        case_study_runs = glob.glob(wild_card)
        loopmate_runs[case_study] = case_study_runs
            
    # collect results
    loopmate_stats = dict()
    for case_study in case_studies:
        loopmate_stats[case_study] = LoopmateStats()
        loopmate_stats[case_study].loopmate_executions = list()
        for work_dir in loopmate_runs[case_study]:
            log_filename = work_dir + "/" + 'loopmate-out.txt'
            if not os.path.exists(log_filename):
                continue
             
            log_file = open(log_filename)
            loopmate_stats[case_study].number_of_repetitions += 1
            curr_execution = LoopmateExecution()
            curr_execution.number_of_evosuite_goals = list()
            loopmate_stats[case_study].loopmate_executions.append(curr_execution)

            escj_loop_inv_counter = 0
            while True:
                line = log_file.readline()
                if not line:
                    if curr_execution.number_of_iterations==0:
                        loopmate_stats[case_study].abnormal_termination += 1
                    break
                elif NORMAL_TERMINATION_VERIFICATION in line:
                    loopmate_stats[case_study].normal_termination += 1
                    loopmate_stats[case_study].stop_criteria_verification += 1
                    #break
                elif NORMAL_TERMINATION_FIXPOINT in line:
                    loopmate_stats[case_study].normal_termination += 1
                    loopmate_stats[case_study].stop_criteria_fixpoint += 1
                    #break
                elif "Total time=" in line:
                    total_time_str = line.split("=")[1]
                    hours = int(total_time_str.split(":")[0])
                    minutes = int(total_time_str.split(":")[1])
                    seconds = float(total_time_str.split(":")[2])
                    
                    total_seconds = hours*60*60 + minutes*60 + seconds
                    curr_execution.execution_time_in_seconds = total_seconds
                elif "Number of iterations=" in line:
                    nrOfIterations = int(line.split("=")[1])
                    curr_execution.number_of_iterations = nrOfIterations
                elif "Inferred Invariants Changed: new iteration is needed" in line:
                    curr_execution.number_of_unguided_iterations += 1
                elif "Evosuite time =" in line:
                    percentage_str = line.split("=")[1]
                    evosuite_time_float= (float(percentage_str.split("%")[0])/100)
                    curr_execution.evosuite_time = evosuite_time_float
                elif "Daikon time   =" in line:
                    percentage_str = line.split("=")[1]
                    daikon_time_float= (float(percentage_str.split("%")[0])/100)
                    curr_execution.daikon_time = daikon_time_float
                elif "ESCJ time      =" in line:
                    percentage_str = line.split("=")[1]
                    escj_time_float= (float(percentage_str.split("%")[0])/100)
                    curr_execution.escj_time = escj_time_float
                
                elif "Total number of test goals:" in line:
                    total_number_of_test_goals = int(line.split(":")[1])
                    curr_execution.number_of_evosuite_goals.append(total_number_of_test_goals)
                    
                elif "ESCJ timeout reached" in line and not "Loop invariant verification:" in line:
                    curr_execution.number_of_escj_timeouts += 1

                
    # print statistics
    print("Loopmate statistics")
    print("===================")

    total_execution_time_in_seconds = 0
    
    success_rates = list()
    failure_rates = list()
    error_rates = list()
    avg_times = list()
    avg_guided_iterations = list()
    avg_unguided_iterations = list()
    avg_evosuite_times = list()
    avg_daikon_times = list()
    avg_escj_times = list()

    
    for case_study in case_studies:
        success = loopmate_stats[case_study].stop_criteria_verification
        failures = loopmate_stats[case_study].stop_criteria_fixpoint
        errors = loopmate_stats[case_study].abnormal_termination
        total = loopmate_stats[case_study].number_of_repetitions
        

        execution_time_list = list()
        iterations_list = list()
        unguided_iteration_list = list()
        evosuite_time_list = list()
        daikon_time_list = list()
        escj_time_list = list()
        escj_timeout_list =list()
        evosuite_target_max_goals_list = list()

        for loopmate_execution in loopmate_stats[case_study].loopmate_executions:
            if loopmate_execution.number_of_iterations!=0:
                execution_time_list.append(loopmate_execution.execution_time_in_seconds)
                total_execution_time_in_seconds += loopmate_execution.execution_time_in_seconds
                iterations_list.append(loopmate_execution.number_of_iterations)
                unguided_iteration_list.append(loopmate_execution.number_of_unguided_iterations)
                evosuite_time_list.append(loopmate_execution.evosuite_time)
                daikon_time_list.append(loopmate_execution.daikon_time)            
                escj_time_list.append(loopmate_execution.escj_time)  
                escj_timeout_list.append(loopmate_execution.number_of_escj_timeouts) 
                evosuite_target_max_goals_list.append(max(loopmate_execution.number_of_evosuite_goals))


        if len(execution_time_list)>0:
            avg_execution_time = datetime.timedelta(seconds=sum(execution_time_list) / len(execution_time_list)) 
            avg_iterations = float(sum(iterations_list))/len(iterations_list)
            avg_unguided_iteration = float(sum(unguided_iteration_list))/len(unguided_iteration_list)
            avg_evosuite_time = float(sum(evosuite_time_list))/len(evosuite_time_list)
            avg_daikon_time = float(sum(daikon_time_list))/len(daikon_time_list) 
            avg_escj_time = float(sum(escj_time_list))/len(escj_time_list) 
            avg_escj_timeouts = float(sum(escj_timeout_list)/len(escj_timeout_list)) 
            avg_max_evosuite_target_goals = float(sum(evosuite_target_max_goals_list)/len(evosuite_target_max_goals_list)) 

            success_rates.append(success/total)
            failure_rates.append(failures/total)
            error_rates.append(errors/total)
            avg_times.append(sum(execution_time_list) / len(execution_time_list))
            avg_guided_iterations.append(avg_iterations)
            avg_evosuite_times.append(avg_evosuite_time)
            avg_daikon_times.append(avg_daikon_time)
            avg_escj_times.append(avg_escj_time)
        else:
            avg_execution_time = float('nan')
            avg_iterations = float('nan')
            avg_unguided_iteration = float('nan')
            avg_evosuite_time = float('nan')
            avg_daikon_time = float('nan')
            avg_escj_time = float('nan')
            avg_escj_timeouts = float('nan')
            avg_max_evosuite_target_goals = float('nan')            
            

        success_rate_str = str(success) + "/" + str(total)
        failure_rate_str = str(failures) + "/" + str(total)
        error_rate_str = str(errors) + "/" + str(total)
        avg_execution_timedelta = avg_execution_time
        avg_guided_iteration_str = str(round(avg_iterations,2))
        avg_unguided_iteration_str = str(round(avg_unguided_iteration,2))
        avg_evosuite_time_str = "{0:.0f}%".format(avg_evosuite_time * 100)
        avg_daikon_time_str = "{0:.0f}%".format(avg_daikon_time * 100)
        avg_escj_time_str = "{0:.0f}%".format(avg_escj_time * 100)
        avg_escj_timeouts_str = str(round(avg_escj_timeouts,2))
        avg_max_evosuite_target_goals_str = str(round(avg_max_evosuite_target_goals,2))
            
        if len(case_study) > 15:
            case_study_col_str = case_study + "\t"
        elif len(case_study) < 8:
            case_study_col_str = case_study + "\t\t\t"
        else:
            case_study_col_str = case_study + "\t\t"
        
        print(case_study_col_str + ": success=" + success_rate_str + \
              "\t failure=" + failure_rate_str + \
              "\t error=" + error_rate_str + \
              "\t avg_time="+ str(avg_execution_timedelta) + \
              "\t avg_guided_iters=" + avg_guided_iteration_str + \
              "\t avg_unguided_iters=" + avg_unguided_iteration_str + \
              "\t avg_evo=" + avg_evosuite_time_str + \
              "\t avg_dkn=" + avg_daikon_time_str + \
              "\t avg_escj=" + avg_escj_time_str  + \
              "\t avg_escj_timeouts=" + avg_escj_timeouts_str + \
              "\t avg_max_evo_goals=" + avg_max_evosuite_target_goals_str)

    print("===================================================================================")
    
    
    total_avg_success_rate = float(sum(success_rates))/len(success_rates) if len(success_rates) > 0 else float('nan')
    total_avg_failure_rate = float(sum(failure_rates))/len(failure_rates) if len(failure_rates) > 0 else float('nan')
    total_avg_error_rate = float(sum(error_rates))/len(error_rates) if len(error_rates) > 0 else float('nan')
    total_avg_exec_time = float(sum(avg_times))/len(avg_times) if len(avg_times) > 0 else float('nan')
    total_avg_iterations = float(sum(avg_guided_iterations))/len(avg_guided_iterations) if len(avg_guided_iterations) > 0 else float('nan')
    total_avg_evosuite_times = float(sum(avg_evosuite_times))/len(avg_evosuite_times) if len(avg_evosuite_times) > 0 else float('nan')
    total_avg_daikon_times = float(sum(avg_daikon_times))/len(avg_daikon_times) if len(avg_daikon_times) > 0 else float('nan')
    total_avg_escj_times = float(sum(avg_escj_times))/len(avg_escj_times) if len(avg_escj_times) > 0 else float('nan')
    

    
    print("AVERAGES : success=" + str(total_avg_success_rate) + \
              "\t failure=" + str(total_avg_failure_rate) + \
              "\t error=" + str(total_avg_error_rate) + \
              "\t avg_time="+ str( datetime.timedelta(seconds=total_avg_exec_time)) + \
              "\t avg_iters=" + str(total_avg_iterations) + \
              "\t avg_evo=" + str(total_avg_evosuite_times) + \
              "\t avg_dkn=" + str(total_avg_daikon_times) + \
              "\t avg_escj=" + str(total_avg_escj_times)  + \
              "\t avg_escj_timeouts=?"  + \
              "\t avg_max_evo_goals=?" )    


    print("===================================================================================")
    print("Total execution time = " + str(datetime.timedelta(seconds=total_execution_time_in_seconds)))