#!/usr/bin/env python
'''
Created on Oct 13, 2012

@author: galeotti
'''

import os
import glob
import datetime
import sys
import math
from dynamate import read_ginpink_invariants, number_of_invariants

NORMAL_TERMINATION_VERIFICATION = "Stop criteria: ESCJ verified source code with loop invariants"
NORMAL_TERMINATION_FIXPOINT = "Stop criteria: daikon inferred invariants did not change"

# ROOT_WORK_DIR_NAME = "/Users/galeotti/workspace/loopmate-python/results"

CWD = os.getcwd()
PYTHON_BIN = "python3.2"
LOOPMATE_PY = CWD + "/scripts/loopmate.py"
#=======================================================================
class LoopmateExecution:
    number_of_iterations = 0
    total_time = 0
    evosuite_time = 0
    daikon_time = 0
    ginpink_time = 0
    escj_time = 0
    execution_time_in_seconds = 0

    initial_number_of_warnings = 0
    daikon_timeout = False
    number_of_inferred_loop_invariants = 0
    final_number_of_warnings = None
    diff_final_warnings = 0

    normal_termination = False
    verification_was_successful = False
    evosuite_restart_threshold = False
    unfinished = False
    
    number_of_verified_ginpink_invariants = 0
    highest_ginpink_wave = 0
    ginpink_candidates = 0
    ginpink_candidates_removed_by_evosuite = 0
    ginpink_candidates_removed_by_escj = 0
#=====================================================================================
def get_curr_number_of_verified_ginpink_invariants(escj_work_dir):
    for verified_ginpink_invs_filename in glob.glob(escj_work_dir + "/" + "verified_ginpink_invs*.txt"):
        ginpink_invariants = read_ginpink_invariants(verified_ginpink_invs_filename)
        count = number_of_invariants(ginpink_invariants)
        return count
    
    return 0

#=====================================================================================

if __name__ == '__main__':
    
    if len(sys.argv) != 2 and  len(sys.argv) != 3 and  len(sys.argv) != 4 :
        print("Usage python stats2.py <results_dir_name> [--ginpink] [--time]")
        print("Example python stats2.py /Users/galeotti/workspace/loopmate-python/results")
        sys.exit()
    
    show_ginpink = False
    show_time = False
    root_work_dir_name = sys.argv[1]
    if (len(sys.argv) > 2 and sys.argv[2] == "--ginpink") or (len(sys.argv) > 3 and sys.argv[3] == "--ginpink"):
        show_ginpink = True

    if (len(sys.argv) > 2 and sys.argv[2] == "--time") or (len(sys.argv) > 3 and sys.argv[3] == "--time"):
        show_time = True
    
    case_studies = list()


    wild_card = root_work_dir_name + "/" + "*" 
    print("Listing " + wild_card) 
    benchmarks = glob.glob(wild_card)
    for benchmark in benchmarks:
        benchmark_id = benchmark.split("/")[-1]
        print ("benchmark: " + benchmark_id)

        wild_card = root_work_dir_name + "/" + benchmark_id + "/*" 
        classnames = glob.glob(wild_card)
        for classname in classnames:
            classname_id = classname.split("/")[-1]
            print ("classname: " + classname_id)

            wild_card = root_work_dir_name + "/" + benchmark_id + "/" + classname_id + "/*" 
            methods = glob.glob(wild_card)
            for method in methods:
                method_id = method.split("/")[-1]
                print("method: " + method_id)
            
                case_studies.append(benchmark_id + "/" + classname_id + "/" + method_id)

    loopmate_runs = dict()
    for case_study in case_studies:
        wild_card = root_work_dir_name + "/" + case_study + "/" + "run_*" 
        print("Listing " + wild_card) 
        case_study_runs = glob.glob(wild_card)
        loopmate_runs[case_study] = case_study_runs
            
    # collect results
    loopmate_executions = dict()
    for case_study in case_studies:
        loopmate_executions[case_study] = list()
        for work_dir in loopmate_runs[case_study]:
            log_filename = work_dir + "/" + 'loopmate-out.txt'
            if not os.path.exists(log_filename):
                continue
            

             
            log_file = open(log_filename)
            
            curr_execution = LoopmateExecution()
            curr_execution.unfinished = True
            loopmate_executions[case_study].append(curr_execution)

            # count number of iterations
            curr_execution.number_of_iterations = 0
            for escj_work_dir in glob.glob(work_dir + "/" + "escj" + "/iter*"):
                if not "." in escj_work_dir: 
                    curr_execution.number_of_iterations += 1
                    curr_number_of_verified_ginpink_invariants = get_curr_number_of_verified_ginpink_invariants(escj_work_dir)
                    curr_number_of_verified_ginpink_invariants = 0
                    if curr_execution.number_of_verified_ginpink_invariants < curr_number_of_verified_ginpink_invariants:
                        curr_execution.number_of_verified_ginpink_invariants = curr_number_of_verified_ginpink_invariants
                else:
                    last_part = escj_work_dir.split(".")[-1]
                    if not last_part.isdigit():
                        curr_execution.number_of_iterations += 1
                        curr_number_of_verified_ginpink_invariants = get_curr_number_of_verified_ginpink_invariants(escj_work_dir)
                        if curr_execution.number_of_verified_ginpink_invariants < curr_number_of_verified_ginpink_invariants:
                            curr_execution.number_of_verified_ginpink_invariants = curr_number_of_verified_ginpink_invariants

            escj_loop_inv_counter = 0
            while True:
                line = log_file.readline()
                if not line:
                    break

                elif "NormalExit" in line or "LoopmateException" in line:
                    curr_execution.unfinished = False
                    curr_execution.normal_termination = True

                elif "LoopmateError" in line or "LoopmateCrash" in line:
                    curr_execution.unfinished = False
                    curr_execution.normal_termination = False

                elif "Highest Ginpink Wave=" in line:
                    number_str = line.split("=")[1]
                    curr_execution.highest_ginpink_wave = int(number_str)

                    # break
                elif "Total time=" in line:
                    total_time_str = line.split("=")[1]
                    hours = int(total_time_str.split(":")[0])
                    minutes = int(total_time_str.split(":")[1])
                    seconds = float(total_time_str.split(":")[2])
                    
                    total_seconds = hours * 60 * 60 + minutes * 60 + seconds
                    curr_execution.execution_time_in_seconds = total_seconds

                elif "Evosuite time" in line and "=" in line:
                    percentage_str = line.split("=")[1]
                    evosuite_time_float = (float(percentage_str.split("%")[0]) / 100)
                    curr_execution.evosuite_time = evosuite_time_float

                elif "Daikon time" in line and "=" in line:
                    percentage_str = line.split("=")[1]
                    daikon_time_float = (float(percentage_str.split("%")[0]) / 100)
                    curr_execution.daikon_time = daikon_time_float

                elif "Ginpink time" in line and "=" in line:
                    percentage_str = line.split("=")[1]
                    ginpink_time_float = (float(percentage_str.split("%")[0]) / 100)
                    curr_execution.ginpink_time = ginpink_time_float

                elif "ESCJ time" in line and "=" in line:
                    percentage_str = line.split("=")[1]
                    escj_time_float = (float(percentage_str.split("%")[0]) / 100)
                    curr_execution.escj_time = escj_time_float
                    
                elif "initial_number_of_warnings=" in line:
                    number_str = line.split("=")[1]
                    initial_number_of_warnings = int(number_str)
                    curr_execution.initial_number_of_warnings = initial_number_of_warnings
                    
                elif "DaikonTimeout" in line:
                    curr_execution.daikon_timeout = True

                elif "number_of_verified_loop_invariants=" in line:
                    number_str = line.split("=")[1]
                    number_of_inferred_loop_invariants = int(number_str)
                    curr_execution.number_of_inferred_loop_invariants = number_of_inferred_loop_invariants
                    
                elif "minimum_number_of_warnings=" in line:
                    number_str = line.split("=")[1]
                    minimum_number_of_warnings = int(number_str)
                    curr_execution.final_number_of_warnings = minimum_number_of_warnings
                    curr_execution.diff_final_warnings = curr_execution.initial_number_of_warnings - minimum_number_of_warnings
                    if minimum_number_of_warnings == 0:
                        curr_execution.verification_was_successful = True
                    
                elif "EvoSuiteRestartLimit" in line:
                    curr_execution.evosuite_restart_threshold = True

                elif "Ginpink Wave:" in line:
                    if "Ginpink new invariants after typechecking=" in line:
                        number_str = line.split("Ginpink new invariants after typechecking=")[1].split("(")[0]
                        wave_ginpink_candidates = int(number_str)
                        curr_execution.ginpink_candidates += wave_ginpink_candidates
                    elif "Ginpink new invariants after testsuite execution=" in line:
                        number_str = line.split("Ginpink new invariants after testsuite execution=")[1].split("(")[0]
                        wave_candidates_after_evosuite = int(number_str)
                        wave_candidates_removed_by_evosuite = wave_ginpink_candidates - wave_candidates_after_evosuite
                        curr_execution.ginpink_candidates_removed_by_evosuite += wave_candidates_removed_by_evosuite
                    elif "Ginpink new invariants after ESC/Java2 removal=" in line:
                        number_str = line.split("Ginpink new invariants after ESC/Java2 removal=")[1].split("(")[0]
                        wave_candidates_after_escj = int(number_str)
                        wave_removed_by_escj = wave_candidates_after_evosuite - wave_candidates_after_escj
                        curr_execution.ginpink_candidates_removed_by_escj += wave_removed_by_escj
                    
                    
    # print statistics
    print("Loopmate statistics")
    print("===================")

    total_execution_time_in_seconds = 0
    avg_times = list()
    maxLength = -1
    for case_study in case_studies:
        if len(case_study) > maxLength:
            maxLength = len(case_study)
      
    partial_verif_count = 0      
    full_verif_count = 0
    normal_exit_count = 0
    avg_reductions_sum = 0
    avg_case_study_partial_verifications = list()
    avg_case_study_full_verifications = list()
    avg_case_study_reductions = list() 
    
    avg_ginpink_candidate_typechecked_list = list()
    avg_candidates_removed_by_evosuite = list()
    avg_candidates_removed_by_escj = list()
    
    final_avg_number_of_iterations_list = list()
    grand_avg_verified_ginpink_invariants_list = list()
    
    avg_evosuite_times = list()
    avg_daikon_times = list()
    avg_ginpink_times = list()
    avg_escj_times = list()

    
    for case_study in sorted(case_studies):

        initial_warnings_list = list()
        inferred_loop_invariant_list = list()
        final_warnings_list = list()
        execution_time_list = list()
        number_of_verification_success = 0
        number_of_partial_verification_success = 0
        number_of_loop_verif_success = 0
        normal_termination_count = 0
        number_of_unfinished_execs = 0
        highest_ginpink_wave_list = list()
        reduction_warnings_list = list()
        ginpink_candidate_list = list()
        ginpink_candidate_evosuite_removal_list = list()
        ginpink_candidate_escj_removal_list = list()
        number_of_iterations_list = list()
        number_of_verified_ginpink_invariants_list = list()
        evosuite_times = list()
        daikon_times = list()
        ginpink_times = list()
        escj_times = list()
        
        for loopmate_execution in loopmate_executions[case_study]:
            
            if loopmate_execution.unfinished == True:
                number_of_unfinished_execs += 1

            if loopmate_execution.normal_termination == True:
                number_of_iterations_list.append(loopmate_execution.number_of_iterations)
                number_of_verified_ginpink_invariants_list.append(loopmate_execution.number_of_verified_ginpink_invariants)
                
                evosuite_times.append(loopmate_execution.evosuite_time)
                daikon_times.append(loopmate_execution.daikon_time)
                ginpink_times.append(loopmate_execution.ginpink_time)
                escj_times.append(loopmate_execution.escj_time)
                
                highest_ginpink_wave_list.append(loopmate_execution.highest_ginpink_wave)
        
                ginpink_candidate_list.append(loopmate_execution.ginpink_candidates)        
                if loopmate_execution.ginpink_candidates > 0:
                    ginpink_candidate_evosuite_removal_rate = float(loopmate_execution.ginpink_candidates_removed_by_evosuite / loopmate_execution.ginpink_candidates)
                    ginpink_candidate_evosuite_removal_list.append(ginpink_candidate_evosuite_removal_rate)
                
                candidates_after_evosuite = loopmate_execution.ginpink_candidates - loopmate_execution.ginpink_candidates_removed_by_evosuite
                if candidates_after_evosuite > 0:
                    ginpink_candidate_escj_removal_rate = float(loopmate_execution.ginpink_candidates_removed_by_escj / candidates_after_evosuite)
                    ginpink_candidate_escj_removal_list.append(ginpink_candidate_escj_removal_rate)
                
                
                normal_termination_count += 1
                if loopmate_execution.verification_was_successful == True:
                    number_of_verification_success += 1
                
                if loopmate_execution.final_number_of_warnings < loopmate_execution.initial_number_of_warnings:
                    number_of_partial_verification_success += 1

                if loopmate_execution.number_of_inferred_loop_invariants > 0:
                    number_of_loop_verif_success += 1
                            
                initial_warnings_list.append(loopmate_execution.initial_number_of_warnings)
                inferred_loop_invariant_list.append(loopmate_execution.number_of_inferred_loop_invariants)
                if  loopmate_execution.final_number_of_warnings != None:
                    final_warnings_list.append(loopmate_execution.final_number_of_warnings)
                else:
                    final_warnings_list.append(loopmate_execution.initial_number_of_warnings)
                    
                reduction_warnings_rate = 1 - float(loopmate_execution.final_number_of_warnings / loopmate_execution.initial_number_of_warnings)
                reduction_warnings_list.append(reduction_warnings_rate)
                    
                execution_time_list.append(loopmate_execution.execution_time_in_seconds)
                total_execution_time_in_seconds += loopmate_execution.execution_time_in_seconds

        if len(execution_time_list) > 0:
            avg_initial_warnings = float(sum(initial_warnings_list) / len(initial_warnings_list))
            avg_inferred_loop_invariants = float(sum(inferred_loop_invariant_list) / len(inferred_loop_invariant_list))
            avg_final_warnings = float(sum(final_warnings_list) / len(final_warnings_list))
            avg_execution_time = datetime.timedelta(seconds=sum(execution_time_list) / len(execution_time_list)) 
            avg_times.append(sum(execution_time_list) / len(execution_time_list))
        else:
            avg_initial_warnings = float('nan')
            avg_inferred_loop_invariants = float('nan')
            avg_final_warnings = float('nan')
            avg_execution_time = float('nan')


        if normal_termination_count > 0:
            avg_verification_success = float(number_of_verification_success / normal_termination_count)
            avg_partial_verification_success = float(number_of_partial_verification_success / normal_termination_count)
            avg_loop_verification_success = float(number_of_loop_verif_success / normal_termination_count)
            avg_highest_ginpink_wave = float(sum(highest_ginpink_wave_list) / normal_termination_count)
            avg_reductions = float(sum(reduction_warnings_list) / len(reduction_warnings_list))
            
            avg_case_study_partial_verifications.append(avg_partial_verification_success)
            avg_case_study_full_verifications.append(avg_verification_success)
            avg_case_study_reductions.append(avg_reductions)


            if len(ginpink_candidate_list) > 0:
                avg_ginpink_candidate_typechecked = float(sum(ginpink_candidate_list) / len(ginpink_candidate_list))
                avg_ginpink_candidate_typechecked_list.append(avg_ginpink_candidate_typechecked)
            else:
                avg_ginpink_candidate_typechecked = float('nan')

            if len(ginpink_candidate_evosuite_removal_list) > 0:
                avg_candidate_evosuite_removal = float(sum(ginpink_candidate_evosuite_removal_list) / len(ginpink_candidate_evosuite_removal_list))
                avg_candidates_removed_by_evosuite.append(avg_candidate_evosuite_removal)
            else: 
                avg_candidate_evosuite_removal = float('nan')
            
            if len(ginpink_candidate_escj_removal_list) > 0:
                avg_candidate_escj_removal = float(sum(ginpink_candidate_escj_removal_list) / len(ginpink_candidate_escj_removal_list))
                avg_candidates_removed_by_escj.append(avg_candidate_escj_removal)
            else:
                avg_candidate_escj_removal = float('nan')
            
        else:
            avg_verification_success = float('nan')
            avg_partial_verification_success = float('nan')
            avg_loop_verification_success = float('nan')
            avg_highest_ginpink_wave = float('nan')
            avg_reductions = float('nan')
            avg_candidate_evosuite_removal = float('nan')
            avg_candidate_escj_removal = float('nan')

        if len(number_of_iterations_list) > 0:
            avg_number_of_iterations_float = float(sum(number_of_iterations_list) / len(number_of_iterations_list))
            final_avg_number_of_iterations_list.append(avg_number_of_iterations_float)
        else:
            avg_number_of_iterations_float = float('nan')

        if len(number_of_verified_ginpink_invariants_list) > 0:
            avg_number_of_verified_ginpink_invariants = float(sum(number_of_verified_ginpink_invariants_list) / len(number_of_verified_ginpink_invariants_list))
            grand_avg_verified_ginpink_invariants_list.append(avg_number_of_verified_ginpink_invariants)
        else:
            avg_number_of_verified_ginpink_invariants = float('nan')


        if len(evosuite_times) > 0:
            avg_evosuite_time_float = float(sum(evosuite_times) / len(evosuite_times))
            avg_evosuite_times.append(avg_evosuite_time_float)
        else:
            avg_evosuite_time_float = float('nan')
            
        if len(daikon_times) > 0:
            avg_daikon_time_float = float(sum(daikon_times) / len(daikon_times))
            avg_daikon_times.append(avg_daikon_time_float)
        else:
            avg_daikon_time_float = float('nan')

        if len(ginpink_times) > 0:
            avg_ginpink_time_float = float(sum(ginpink_times) / len(ginpink_times))
            avg_ginpink_times.append(avg_ginpink_time_float)
        else:
            avg_ginpink_time_float = float('nan')
            
        if len(escj_times) > 0:
            avg_escj_time_float = float(sum(escj_times) / len(escj_times))
            avg_escj_times.append(avg_escj_time_float)
        else:
            avg_escj_time_float = float('nan')

        avg_initial_warnings_str = str(round(avg_initial_warnings, 2))
        avg_inferred_loop_invariants_str = str(round(avg_inferred_loop_invariants, 2))
        avg_final_warnings_str = str(round(avg_final_warnings, 2))
        if avg_ginpink_candidate_typechecked != float('nan'):
            avg_ginpink_candidate_typechecked_str = str(round(avg_ginpink_candidate_typechecked, 2))
        else:
            avg_ginpink_candidate_typechecked_str = str(float('nan'))
            
        if avg_number_of_iterations_float != float('nan'):
            avg_number_of_iterations_str = str(round(avg_number_of_iterations_float, 2))
        else:
            avg_number_of_iterations_str = str(float('nan'))
            
        if avg_number_of_verified_ginpink_invariants != float('nan'):
            avg_number_of_verified_ginpink_invariants_str = str(round(avg_number_of_verified_ginpink_invariants, 2))
        else:
            avg_number_of_verified_ginpink_invariants_str = str(float('nan'))
            
        avg_partial_verification_success_str = "{0:.0f}%".format(avg_partial_verification_success * 100)
        avg_verification_success_str = "{0:.0f}%".format(avg_verification_success * 100)
        avg_loop_verification_success_str = "{0:.0f}%".format(avg_loop_verification_success * 100)
        avg_highest_ginpink_wave_str = str(round(avg_highest_ginpink_wave, 2))
        avg_candidate_evosuite_removal_str = "{0:.0f}%".format(avg_candidate_evosuite_removal * 100)
        avg_candidates_removed_by_escj_str = "{0:.0f}%".format(avg_candidate_escj_removal * 100)
        avg_reduction_str = "{0:.0f}%".format(avg_reductions * 100)

        avg_evosuite_times_str = "{0:.0f}%".format(avg_evosuite_time_float * 100)
        avg_daikon_times_str = "{0:.0f}%".format(avg_daikon_time_float * 100)
        avg_ginpink_times_str = "{0:.0f}%".format(avg_ginpink_time_float * 100)
        avg_escj_times_str = "{0:.0f}%".format(avg_escj_time_float * 100)
        
        avg_execution_timedelta = avg_execution_time
            
        line = case_study.ljust(maxLength) + ":" + \
               "\t" + "warns=" + avg_initial_warnings_str + \
                "\t" + "final=" + avg_final_warnings_str + \
                "\t" + "reduction=" + avg_reduction_str + \
                "\t" + "loop_invs=" + avg_inferred_loop_invariants_str + \
                "\t" + "iterations=" + avg_number_of_iterations_str
                
        if show_ginpink:
            line += "\t\t" + "ginpink_wave=" + avg_highest_ginpink_wave_str + \
                "\t" + "ginpink_candidates=" + avg_ginpink_candidate_typechecked_str + \
                "\t" + "ginpink_evosuite=" + avg_candidate_evosuite_removal_str + \
                "\t" + "ginpink_escj=" + avg_candidates_removed_by_escj_str + \
                "\t" + "ginpink_loop_invs=" + avg_number_of_verified_ginpink_invariants_str

        if show_time:
            line += "\t\t" + "time=" + str(avg_execution_timedelta) + \
                "\t" + "evosuite_time=" + str(avg_evosuite_times_str) + \
                "\t" + "daikon_time=" + str(avg_daikon_times_str) + \
                "\t" + "ginpink_time=" + str(avg_ginpink_times_str) + \
                "\t" + "escj_time=" + str(avg_escj_times_str) 

        line += "\t" + "partial_verif=" + str(avg_partial_verification_success_str) + \
               "\t" + "full_verif=" + str(avg_verification_success_str) + \
               "\t" + "normal_runs=" + str(normal_termination_count) + "/" + str(len(loopmate_executions[case_study])) + \
               "  " + "(" + str(number_of_unfinished_execs) + ")"
        
        print (line)
        
        partial_verif_count += number_of_partial_verification_success
        full_verif_count += number_of_verification_success
        normal_exit_count += normal_termination_count
        avg_reductions_sum += avg_reductions
                          

    print("===================================================================================")


    if len(avg_case_study_partial_verifications) > 0:

        avg_partial_verif = float(sum(avg_case_study_partial_verifications) / len(avg_case_study_partial_verifications))
        avg_full_verif = float(sum(avg_case_study_full_verifications) / len(avg_case_study_full_verifications))
        avg_reduction_verif = float(sum(avg_case_study_reductions) / len(avg_case_study_reductions))

        if len(avg_ginpink_candidate_typechecked_list) > 0:
            final_avg_ginpink_candidate_typechecked_float = float(sum(avg_ginpink_candidate_typechecked_list) / len(avg_ginpink_candidate_typechecked_list))
        else:
            final_avg_ginpink_candidate_typechecked_float = float('nan')
        
        if len(avg_candidates_removed_by_evosuite) > 0:
            avg_evosuite_removal = float(sum(avg_candidates_removed_by_evosuite) / len(avg_candidates_removed_by_evosuite))
        else:
            avg_evosuite_removal = float('nan')

        if len(avg_candidates_removed_by_escj) > 0:
            avg_escj_removal = float(sum(avg_candidates_removed_by_escj) / len(avg_candidates_removed_by_escj))
        else:
            avg_escj_removal = float('nan')

        if len(avg_evosuite_times) > 0:
            final_avg_evosuite_time = float(sum(avg_evosuite_times) / len(avg_evosuite_times))
        else:
            final_avg_evosuite_time = float('nan')

        if len(avg_daikon_times) > 0:
            final_avg_daikon_time = float(sum(avg_daikon_times) / len(avg_daikon_times))
        else:
            final_avg_daikon_time = float('nan')

        if len(avg_ginpink_times) > 0:
            final_avg_ginpink_time = float(sum(avg_ginpink_times) / len(avg_ginpink_times))
        else:
            final_avg_ginpink_time = float('nan')

        if len(avg_escj_times) > 0:
            final_avg_escj_time = float(sum(avg_escj_times) / len(avg_escj_times))
        else:
            final_avg_escj_time = float('nan')

        if len(final_avg_number_of_iterations_list) > 0:
            final_avg_number_of_iteration_float = float(sum(final_avg_number_of_iterations_list) / len(final_avg_number_of_iterations_list))
            final_avg_number_of_iteration_str = str(round(final_avg_number_of_iteration_float, 2))
        else:
            final_avg_number_of_iterations_float = float('nan')
            final_avg_number_of_iteration_str = str(float('nan'))

        if len(grand_avg_verified_ginpink_invariants_list)>0:
            grand_avg_verified_ginpink_invariants_float = float(sum(grand_avg_verified_ginpink_invariants_list)/len(grand_avg_verified_ginpink_invariants_list))
            grand_avg_verified_ginpink_invariants_str = str(round(grand_avg_verified_ginpink_invariants_float,2))
        else:
            grand_avg_verified_ginpink_invariants_str = str(float('nan'))


        avg_partial_verif_str = "{0:.0f}%".format(avg_partial_verif * 100)
        avg_full_verif_str = "{0:.0f}%".format(avg_full_verif * 100)
        avg_recuction_verif_str = "{0:.0f}%".format(avg_reduction_verif * 100)
        avg_evosuite_removal_str = "{0:.0f}%".format(avg_evosuite_removal * 100)
        avg_escj_removal_str = "{0:.0f}%".format(avg_escj_removal * 100)
        final_avg_evosuite_time_str = "{0:.0f}%".format(final_avg_evosuite_time * 100)
        final_avg_daikon_time_str = "{0:.0f}%".format(final_avg_daikon_time * 100)
        final_avg_ginpink_time_str = "{0:.0f}%".format(final_avg_ginpink_time * 100)
        final_avg_escj_time_str = "{0:.0f}%".format(final_avg_escj_time * 100)
        final_avg_ginpink_candidate_typechecked_str = str(final_avg_ginpink_candidate_typechecked_float)

        print("normal_runs=" + str(normal_exit_count))
        print("avg_number_of_iterations=" + final_avg_number_of_iteration_str)
        print("avg_partial_verification=" + avg_partial_verif_str)
        print("avg_full_verification=" + avg_full_verif_str)   
        print("avg_reduction_of_warnings=" + avg_recuction_verif_str)
        print("grand_avg_verified_ginpink_invariants=" + grand_avg_verified_ginpink_invariants_str)
        print("final_avg_ginpink_candidate_typechecked=" + final_avg_ginpink_candidate_typechecked_str)
        print("avg_ginpink_candidate_evosuite_removal=" + avg_evosuite_removal_str)
        print("avg_ginpink_candidate_escj_removal=" + avg_escj_removal_str)
        print("avg_evosuite_time=" + final_avg_evosuite_time_str)
        print("avg_daikon_time=" + final_avg_daikon_time_str)
        print("avg_ginpink_time=" + final_avg_ginpink_time_str)
        print("avg_escj_time=" + final_avg_escj_time_str)
        
    print("===================================================================================")
    print("Total execution time = " + str(datetime.timedelta(seconds=total_execution_time_in_seconds)))
