package org.ginpink;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.RecognitionException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

public final class ReadMutations {

	private static final String GINPINK_INV_FILE = "ginpink_inv_file";
	private static final String CLASS_NAME = "class_name";
	private static final String METHOD_NAME = "method_name";

	public ReadMutations(String className, String methodName) {
		this.className = className;
		this.methodName = methodName;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {

		Options options = new Options();
		options.addOption(CLASS_NAME, true, "target class name");
		options.addOption(METHOD_NAME, true, "target method name");
		options.addOption(GINPINK_INV_FILE, true,
				"output ginpink invariants file name");

		CommandLineParser parser = new GnuParser();
		try {
			CommandLine commandLine = parser.parse(options, args);
			List<String> mutation_filenames = commandLine.getArgList();
			if (mutation_filenames.size() == 0) {
				throw new ParseException(
						"Usage [options] <mutationsfile1> <mutationsfile2>");
			}

			if (!commandLine.hasOption(GINPINK_INV_FILE)) {
				throw new ParseException("Missing mandatory argument: "
						+ GINPINK_INV_FILE);
			}
			String invariantsFilename = commandLine
					.getOptionValue(GINPINK_INV_FILE);

			if (!commandLine.hasOption(CLASS_NAME)) {
				throw new ParseException("Missing mandatory argument: "
						+ CLASS_NAME);
			}
			String className = commandLine.getOptionValue(CLASS_NAME);

			if (!commandLine.hasOption(METHOD_NAME)) {
				throw new ParseException("Missing mandatory argument: "
						+ METHOD_NAME);
			}
			String methodName = commandLine.getOptionValue(METHOD_NAME);

			ReadMutations reader = new ReadMutations(className, methodName);
			reader.readMutations(mutation_filenames, invariantsFilename);

		} catch (ParseException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (RecognitionException e) {
			throw new RuntimeException(e);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
	}

	private final String className;
	private final String methodName;

	public void readMutations(List<String> mutationsFilenames,
			String invFilename) throws IOException, RecognitionException,
			ClassNotFoundException, NoSuchMethodException {

		GinPinkChecker checker = new GinPinkChecker(className, methodName);
		for (String mutationsFilename : mutationsFilenames) {
			System.out.println("Parsing mutations filename "
					+ mutationsFilename);
			List<Invariant> mutations = checker
					.parseMutations(mutationsFilename);
			System.out.println("Number of parsed mutations " + mutations.size()
					+ " from file " + mutationsFilename);
		}

		System.out.println("Typechecking parsed mutations");
		checker.checkTyping();
		Set<String> loopinv_methods = checker.getLoopInvariantMethodNames();
		System.out.println("Found " + loopinv_methods.size() + " methods");
		for (String loopinv_method : loopinv_methods) {
			List<Invariant> typechecked_exprs = checker
					.getInvariants(loopinv_method);
			System.out.println("Method " + loopinv_method + " -> "
					+ typechecked_exprs.size() + " typechecked invariants");
		}
		
		System.out.println("Writting typechecked invariants to file "
				+ invFilename);
		checker.writeInvariants(invFilename);
		System.out.println("ReadMutations Done!");

	}
}
