package org.ginpink.daikon;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.regex.Pattern;

import daikon.FileIO;
import daikon.PptMap;
import daikon.PptTopLevel;
import daikon.ValueTuple;
import daikon.VarInfo;

public final class GinPinkProcessor extends FileIO.Processor {

	private LinkedList<GinPinkProcessorObserver> observers = new LinkedList<GinPinkProcessorObserver>();
	private String ppt_omit_pattern;
	private String ppt_select_pattern;

	public boolean registerObserver(GinPinkProcessorObserver snapshotObserver) {
		return observers.add(snapshotObserver);
	}

	public boolean unregisterObserver(GinPinkProcessorObserver snapshotObserver) {
		return observers.remove(snapshotObserver);
	}

	@Override
	public void process_sample(PptMap all_ppts, PptTopLevel ppt, ValueTuple vt,
			Integer nonce) {

		if (ppt_omit_pattern != null) {
			Pattern pattern = Pattern.compile(ppt_omit_pattern);
			if (pattern.matcher(ppt.name()).find()) {
				return;
			}
		}

		if (ppt_select_pattern != null) {
			Pattern pattern = Pattern.compile(ppt_select_pattern);
			if (!pattern.matcher(ppt.name()).find()) {
				return;
			}
		}

		final VarInfo[] var_infos = ppt.var_infos;

		String ppt_name = ppt.name();
		String[] classNameAndMethodName = getClassNameAndMethodName(ppt_name);
		String className = classNameAndMethodName[0];
		String methodName = classNameAndMethodName[1];

		SnapshotBuilder snapshotBuilder = new SnapshotBuilder();
		Snapshot snapshot = snapshotBuilder.buildSnapshot(className,
				methodName, var_infos, vt);
		notifyNewSnapshot(snapshot);

	}

	private static String[] getClassNameAndMethodName(String ppt_name) {
		String[] split_parts = ppt_name.split("\\(");
		String[] className_methodName_parts = split_parts[0].split("\\.");

		String methodName = className_methodName_parts[className_methodName_parts.length - 1];
		String[] className_parts = Arrays.copyOfRange(
				className_methodName_parts, 0,
				className_methodName_parts.length - 1);

		String className = join(className_parts);

		return new String[] { className, methodName };
	}

	private static String join(String[] className_parts) {
		StringBuffer buff = new StringBuffer();
		for (int i = 0; i < className_parts.length; i++) {
			if (buff.length() > 0) {
				buff.append(".");
			}
			buff.append(className_parts[i]);
		}
		return buff.toString();
	}

	private void notifyNewSnapshot(Snapshot snapshot) {
		for (GinPinkProcessorObserver observer : observers) {
			observer.notifyNewSnapshot(snapshot);
		}
	}

	public String getPptOmitPattern() {
		return ppt_omit_pattern;
	}

	public void setPptOmitPattern(String regex) {
		this.ppt_omit_pattern = regex;
	}

	public String getPptSelectPattern() {
		return ppt_select_pattern;
	}

	public void setPptSelectPattern(String regex) {
		this.ppt_select_pattern = regex;
	}

}
