package org.ginpink.daikon;

import java.util.HashMap;
import java.util.Set;

public final class Snapshot {

	private HashMap<String, SnapshotValue> variables = new HashMap<String, SnapshotValue>();
	private HashMap<String, SnapshotValue> classFields = new HashMap<String, SnapshotValue>();

	private SnapshotIntValue.SnapshotIntValueBuilder snapshotIntValueBuilder = new SnapshotIntValue.SnapshotIntValueBuilder();
	private SnapshotNonNullReference.SnapshotNonNullReferenceBuilder nonNullReferenceBuilder = new SnapshotNonNullReference.SnapshotNonNullReferenceBuilder();
	private SnapshotLongValue.SnapshotLongValueBuilder snapshotLongValueBuilder = new SnapshotLongValue.SnapshotLongValueBuilder();

	private final String className;
	private final String methodName;

	public Snapshot(String className, String methodName) {
		this.className = className;
		this.methodName = methodName;
	}

	public boolean isVariable(String varName) {
		return variables.containsKey(varName);
	}

	public SnapshotIntValue getIntValue(int intValue) {
		SnapshotIntValue snapshotIntValue = snapshotIntValueBuilder
				.getSnapshotIntValue(intValue);
		return snapshotIntValue;
	}

	public SnapshotNonNullReference getNonNullReference(int hashCodeValue) {
		SnapshotNonNullReference snapshotReference = nonNullReferenceBuilder
				.getSnapshotReference(hashCodeValue);
		return snapshotReference;
	}

	public SnapshotNullReference getNullReference() {
		return SnapshotNullReference.NULL_REFERENCE;
	}

	public void setVariable(String varName, SnapshotValue snapshotValue) {
		variables.put(varName, snapshotValue);
	}

	public void setClassField(String classFieldName, SnapshotValue snapshotValue) {
		classFields.put(classFieldName, snapshotValue);
	}

	public SnapshotNonNullReference getReference(String referenceVariableName) {
		SnapshotValue snapshotValue = variables.get(referenceVariableName);
		SnapshotNonNullReference snapshotReference = (SnapshotNonNullReference) snapshotValue;
		return snapshotReference;
	}

	public SnapshotValue getInstanceField(SnapshotNonNullReference ref,
			String fieldName) {
		return ref.getFieldValue(fieldName);
	}

	public void setInstanceField(String fieldName,
			SnapshotNonNullReference curr, SnapshotValue snapshotValue) {
		curr.setFieldValue(fieldName, snapshotValue);
	}

	public SnapshotValue getVariable(String varName) {
		return variables.get(varName);
	}

	public boolean isStaticField(String staticFieldName) {
		return classFields.containsKey(staticFieldName);
	}

	public SnapshotValue getClassField(String staticFieldName) {
		return classFields.get(staticFieldName);
	}

	public SnapshotValue evaluateDotExpression(String dotExpression) {

		String[] navigation_path = dotExpression.split("\\.");
		SnapshotValue leftmostReceiver = null;

		int i = navigation_path.length - 1;
		for (; i >= 0; i--) {
			String prefixExpr = join(navigation_path, 0, i);
			if (this.isStaticField(prefixExpr)) {
				leftmostReceiver = this.getClassField(prefixExpr);
				break;
			} else if (this.isVariable(prefixExpr)) {
				leftmostReceiver = this.getVariable(prefixExpr);
				break;
			}
		}

		if (leftmostReceiver == null)
			return null;

		if (i == navigation_path.length - 1) {
			return leftmostReceiver;
		}

		SnapshotValue curr_value = leftmostReceiver;
		for (int j = i + 1; j < navigation_path.length; j++) {
			SnapshotNonNullReference curr_ref = (SnapshotNonNullReference) curr_value;
			String currFieldName = navigation_path[j];
			curr_value = curr_ref.getFieldValue(currFieldName);
		}

		return curr_value;
	}

	private static String join(String[] navigation_path, int fromIndex,
			int toIndex) {
		StringBuffer buff = new StringBuffer();
		for (int i = fromIndex; i <= toIndex; i++) {
			if (buff.length() > 0) {
				buff.append(".");
			}
			buff.append(navigation_path[i]);
		}
		return buff.toString();
	}

	public Set<String> parameters() {
		return this.variables.keySet();
	}

	public Set<String> class_fields() {
		return this.classFields.keySet();
	}

	public SnapshotBooleanValue getBooleanValue(boolean booleanValue) {
		return SnapshotBooleanValue.getSnapshotBooleanValue(booleanValue);
	}

	public String getClassName() {
		return className;
	}

	public String getMethodName() {
		return methodName;
	}

	public SnapshotValue getLongValue(long longValue) {
		SnapshotLongValue snapshotLongValue = snapshotLongValueBuilder
				.getSnapshotLongValue(longValue);
		return snapshotLongValue;
	}

}
