package org.ginpink.expr;

public final class JBinaryOperation extends JOperation {

	private final JExpr left;
	private final JOperator operator;
	private final JExpr right;

	public JBinaryOperation(JExpr left, JOperator binary_op, JExpr right) {
		this.left = left;
		this.operator = binary_op;
		this.right = right;
	}

	@Override
	public <T> T accept(JExprVisitor<T> v) {
		return v.visit(this);
	}

	public JExpr getLeft() {
		return left;
	}

	public JOperator getOperator() {
		return operator;
	}

	public JExpr getRight() {
		return right;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((left == null) ? 0 : left.hashCode());
		result = prime * result
				+ ((operator == null) ? 0 : operator.hashCode());
		result = prime * result + ((right == null) ? 0 : right.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JBinaryOperation other = (JBinaryOperation) obj;
		if (left == null) {
			if (other.left != null)
				return false;
		} else if (!left.equals(other.left))
			return false;
		if (operator != other.operator)
			return false;
		if (right == null) {
			if (other.right != null)
				return false;
		} else if (!right.equals(other.right))
			return false;
		return true;
	}

}
