package org.ginpink.expr;

import java.util.Arrays;
import java.util.LinkedList;

public final class JInstanceMethodCall extends JExpr {

	private final String method_name;

	private final JExpr receiver;

	private final LinkedList<JExpr> arguments;

	public JInstanceMethodCall(JExpr receiver,
			String method_name, JExpr... args) {
		this.method_name = method_name;
		this.receiver = receiver;
		this.arguments = new LinkedList<JExpr>(Arrays.asList(args));
	}

	@Override
	public <T> T accept(JExprVisitor<T> v) {
		return v.visit(this);
	}

	public String getMethodName() {
		return method_name;
	}

	public JExpr getReceiver() {
		return receiver;
	}

	public LinkedList<JExpr> getArguments() {
		return arguments;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((arguments == null) ? 0 : arguments.hashCode());
		result = prime * result
				+ ((method_name == null) ? 0 : method_name.hashCode());
		result = prime * result
				+ ((receiver == null) ? 0 : receiver.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		JInstanceMethodCall other = (JInstanceMethodCall) obj;
		if (arguments == null) {
			if (other.arguments != null)
				return false;
		} else if (!arguments.equals(other.arguments))
			return false;
		if (method_name == null) {
			if (other.method_name != null)
				return false;
		} else if (!method_name.equals(other.method_name))
			return false;
		if (receiver == null) {
			if (other.receiver != null)
				return false;
		} else if (!receiver.equals(other.receiver))
			return false;
		return true;
	}
}
