package org.ginpink.expr;

public final class JType {

	private final boolean isPrimitive;

	private final Class<?> reflect_type;

	public boolean isPrimitive() {
		return isPrimitive;
	}

	public boolean isNullType() {
		return this == NULL_TYPE;
	}

	private JType(boolean isPrimitive) {
		this.isPrimitive = isPrimitive;
		reflect_type = null;
	}

	public JType(Class<?> reflect_type) {
		this.reflect_type = reflect_type;
		this.isPrimitive = isPrimitive(reflect_type);
	}

	public static JType NULL_TYPE = new JType(false);

	private static boolean isPrimitive(Class<?> clazz) {
		if (clazz.equals(int.class)) {
			return true;
		}
		if (clazz.equals(long.class)) {
			return true;
		}
		if (clazz.equals(boolean.class)) {
			return true;
		}
		return false;
	}

	public boolean isInt() {
		if (isNullType())
			return false;

		return reflect_type.equals(int.class);
	}

	public boolean isBoolean() {
		if (isNullType())
			return false;

		return reflect_type.equals(boolean.class);
	}

	public boolean equals(Object o) {
		if (o == null)
			return false;

		if (this == o)
			return true;

		if (o instanceof JType) {
			JType that = (JType) o;
			return this.isPrimitive == that.isPrimitive
					&& this.reflect_type == null ? that.reflect_type == null
					: this.reflect_type.equals(that.reflect_type);
		}

		return false;
	}

	public int hashCode() {
		return (this.reflect_type == null ? 0 : this.reflect_type.hashCode())
				+ (this.isPrimitive ? 0 : 1);
	}

	public Class<?> getJavaClass() {
		return reflect_type;
	}

	public boolean isArray() {
		if (isNullType())
			return false;

		return reflect_type.isArray();
	}

	public Class<?> getComponentType() {
		return reflect_type.getComponentType();
	}

	public boolean isLong() {
		if (isNullType())
			return false;

		return reflect_type.equals(long.class);
	}
}
