package org.ginpink.parser;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.ginpink.Invariant;
import org.ginpink.LoopInvariantMethod;
import org.ginpink.expr.JExpr;
import org.ginpink.expr.JExprPrinter;

public final class InvariantFileWriter {

	private static final String EXEC_JML = "<--exec|JML-->";
	private static final String SEPARATOR_LINE = "===========================================================================";
	private static final String LINE_SEPARATOR = System
			.getProperty("line.separator");

	private final List<LoopInvariantMethod> loop_invariant_methods;

	public InvariantFileWriter(List<LoopInvariantMethod> loop_invariant_methods) {
		this.loop_invariant_methods = new LinkedList<LoopInvariantMethod>(
				loop_invariant_methods);
	}

	public void writeFile(String filename) throws IOException {
		writeFile(filename, false);
	}

	public void writeFile(String filename, boolean writeDiscardedInvariants)
			throws IOException {
		BufferedWriter out = new BufferedWriter(new FileWriter(filename));
		Collections.<LoopInvariantMethod> sort(this.loop_invariant_methods);
		for (LoopInvariantMethod loopInvariantMethod : this.loop_invariant_methods) {
			writeLoopHeader(out, loopInvariantMethod.getReflectMethod());
			List<Invariant> invariants;
			if (writeDiscardedInvariants) {
				invariants = loopInvariantMethod.getDiscardedInvariants();
			} else {
				invariants = loopInvariantMethod.getInvariants();
			}
			writeInvariants(out, invariants);
		}
	}

	private static void writeInvariants(BufferedWriter out,
			List<Invariant> invariants) throws IOException {
		for (Invariant invariant : invariants) {
			String java_string = invariant.getJavaString();
			String jml_string = invariant.getJmlString();
			String inv_str = java_string + EXEC_JML + jml_string;
			println(out, inv_str);
		}
	}

	private static void writeLoopHeader(BufferedWriter out, Method method)
			throws IOException {
		println(out, SEPARATOR_LINE);
		Class<?>[] parameterTypes = method.getParameterTypes();
		StringBuffer paramTypesBuff = new StringBuffer();
		for (Class<?> parameterType : parameterTypes) {
			if (paramTypesBuff.length() > 0) {
				paramTypesBuff.append(", ");
			}
			String canonicalName = parameterType.getCanonicalName();
			paramTypesBuff.append(canonicalName);
		}

		String loopHeader = String.format("%s.%s(%s):::ENTER", method
				.getDeclaringClass().getCanonicalName(), method.getName(),
				paramTypesBuff.toString());

		println(out, loopHeader);
	}

	private static void println(BufferedWriter out, String str)
			throws IOException {
		out.write(str);
		out.write(LINE_SEPARATOR);
		out.flush();
	}
}
