package org.ginpink.util;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.LinkedList;

import org.ginpink.expr.JExpr;
import org.ginpink.typecheck.ExprTypeChecker;

public abstract class Reflection {

	public static Method getDeclaredMethod(Class<?> clazz, String methodName,
			Class<?>[] concrete_parameter_types) throws NoSuchMethodException {

		LinkedList<Method> matchingMethods = new LinkedList<Method>();
		for (Method declaredMethod : clazz.getDeclaredMethods()) {
			Class<?>[] formal_parameter_types = declaredMethod
					.getParameterTypes();
			if (declaredMethod.getName().equals(methodName)) {
				boolean checkArgumentTypes = ExprTypeChecker
						.checkArgumentTypes(concrete_parameter_types,
								formal_parameter_types);
				if (checkArgumentTypes) {
					matchingMethods.add(declaredMethod);

				}
			}
		}

		if (matchingMethods.isEmpty()) {
			throw new NoSuchMethodException();
		}

		if (matchingMethods.size() > 1) {
			throw new RuntimeException();
		}

		return matchingMethods.getFirst();
	}

	public static Class<?>[] getParameterTypes(
			LinkedList<JExpr> argumentExprList, Object[] concreteArguments) {
		Class<?>[] parameterTypes = new Class<?>[concreteArguments.length];
		for (int i = 0; i < concreteArguments.length; i++) {
			if (concreteArguments[i] != null) {
				Class<? extends Object> concreteArgumentType = concreteArguments[i]
						.getClass();
				if (argumentExprList.get(i).isPrimitive()) {
					if (concreteArgumentType.equals(Integer.class)) {
						parameterTypes[i] = int.class;
					} else if (concreteArgumentType.equals(Long.class)) {
						parameterTypes[i] = long.class;
					} else if (concreteArgumentType.equals(Boolean.class)) {
						parameterTypes[i] = boolean.class;
					} else {
						throw new RuntimeException(
								"Unhandled primitive type for wrapped type "
										+ concreteArgumentType
												.getCanonicalName());
					}
				} else {
					parameterTypes[i] = concreteArguments[i].getClass();
				}
			}
		}
		return parameterTypes;
	}

	public static Object invokeMethod(final Class<?> clazz, String method_name,
			Object concrete_receiver, Object[] concreteArguments,
			Class<?>[] parameterTypes) throws InvocationTargetException {
		try {

			Method method = Reflection.getDeclaredMethod(clazz, method_name,
					parameterTypes);

			boolean isAccessible = method.isAccessible();
			method.setAccessible(true);
			Object ret_val = method
					.invoke(concrete_receiver, concreteArguments);
			method.setAccessible(isAccessible);

			return ret_val;

		} catch (SecurityException e1) {
			throw new RuntimeException(e1);

		} catch (NoSuchMethodException e1) {
			throw new RuntimeException("The method " + method_name
					+ " with types " + Arrays.toString(parameterTypes)
					+ " was not found", e1);
		} catch (IllegalArgumentException e1) {
			throw new RuntimeException(e1);
		} catch (IllegalAccessException e1) {
			throw new RuntimeException(e1);
		}
	}

}
